import collections
import gc
import warnings

import numpy as np
import scipy.integrate

from .motion_models import constant_velocity_time_to_arrive_1d, approx_two_point_time_to_arrive, constant_velocity_time_to_arrive
from .utility import progress_bar


# SimulationResult object to hold simulation results
_result_fields = [
    "attack_cum_prob",  # F x PHI x T
    "attack_cum_poss",  # F x PHI x T
    "attack_prob_density",  # F x PHI x T
    "attack_poss_density",  # F x PHI x T
    "defense_cum_prob",  # F x PHI x T
    "defense_cum_poss",  # F x PHI x T
    "defense_prob_density",  # F x PHI x T
    "defense_poss_density",  # F x PHI x T

    "cum_p0",  # F x PHI x T
    "p0_density",  # F x PHI x T

    "player_cum_prob",  # F x P x PHI x T
    "player_cum_poss",  # F x P x PHI x T
    "player_prob_density",  # F x P x PHI x T
    "player_poss_density",  # F x P x PHI x T

    "phi_grid",  # PHI
    "r_grid",  # T
    "x_grid",  # F x PHI x T
    "y_grid",  # F x PHI x T
]
ALL_OPTIONAL_FIELDS_TO_RETURN = tuple([field for field in _result_fields if field not in ["phi_grid", "r_grid", "x_grid", "y_grid"]])
SimulationResult = collections.namedtuple("SimulationResult", _result_fields, defaults=[None] * len(_result_fields))

# Default model parameters
_DEFAULT_A_MAX = 10.659091365334193
_DEFAULT_B0 = -4.565680899844368
_DEFAULT_B1 = -188.74468208593532
_DEFAULT_EXCLUDE_PASSER = True
_DEFAULT_FACTOR = 5.077423030272923
_DEFAULT_FACTOR2 = 1.0063028450754512
_DEFAULT_INERTIAL_SECONDS = 1.1043767821571149
_DEFAULT_KEEP_INERTIAL_VELOCITY = True
_DEFAULT_N_V0 = 13.751097117532021
_DEFAULT_NORMALIZE = False
_DEFAULT_PASS_START_LOCATION_OFFSET = -1.5245340256423476
_DEFAULT_PLAYER_VELOCITY = 34.6836072667285
_DEFAULT_RADIAL_GRIDSIZE = 5.034759576558597
_DEFAULT_RESPECT_OFFSIDE = False
_DEFAULT_TIME_OFFSET_BALL = -0.4384754490159207
_DEFAULT_TOL_DISTANCE = 9.986761680941445
_DEFAULT_USE_APPROX_TWO_POINT = True
_DEFAULT_USE_EFFICIENT_SIGMOID = True
_DEFAULT_USE_FIXED_V0 = True
_DEFAULT_USE_MAX = True
_DEFAULT_USE_POSS = True
_DEFAULT_V0_MAX = 42.18118275402132
_DEFAULT_V0_MIN = 8.886015553615485
_DEFAULT_V0_PROB_AGGREGATION_MODE = "mean"
_DEFAULT_V_MAX = 19.85563874348074


def _approximate_sigmoid(x):
    """
    Computational efficient sigmoid approximation.

    >>> _approximate_sigmoid(np.array([-1, 0, 1])), 1 / (1 + np.exp(-np.array([-1, 0, 1])))
    (array([0.25, 0.5 , 0.75]), array([0.26894142, 0.5       , 0.73105858]))
    """
    return 0.5 * (x / (1 + np.abs(x)) + 1)


def _sigmoid(x):
    """
    >>> _sigmoid(np.array([-1, 0, 1]))
    array([0.26894142, 0.5       , 0.73105858])
    """
    return 1 / (1 + np.exp(-x))


def _assert_matrix_consistency(PLAYER_POS, BALL_POS, phi_grid, v0_grid, passer_team, team_list, players=None, passers=None):
    F = PLAYER_POS.shape[0]
    assert F == BALL_POS.shape[0], f"Dimension F is {F} (from PLAYER_POS: {PLAYER_POS.shape}), but BALL_POS shape is {BALL_POS.shape}"
    assert F == phi_grid.shape[0], f"Dimension F is {F} (from PLAYER_POS: {PLAYER_POS.shape}), but phi_grid shape is {phi_grid.shape}"
    assert F == v0_grid.shape[0], f"Dimension F is {F} (from PLAYER_POS: {PLAYER_POS.shape}), but v0_grid shape is {v0_grid.shape}"
    assert F == passer_team.shape[0], f"Dimension F is {F} (from PLAYER_POS: {PLAYER_POS.shape}), but passer_team shape is {passer_team.shape}"
    P = PLAYER_POS.shape[1]
    assert P == team_list.shape[0], f"Dimension P is {P} (from PLAYER_POS: {PLAYER_POS.shape}), but team_list shape is {team_list.shape}"
    assert PLAYER_POS.shape[2] >= 4  # >= or = ?
    assert BALL_POS.shape[1] >= 2  # ...
    if passers is not None:
        assert F == passers.shape[0], f"Dimension F is {F} (from PLAYER_POS: {PLAYER_POS.shape}), but passers shape is {passers.shape}"
        assert P == players.shape[0], f"Dimension P is {P} (from PLAYER_POS: {PLAYER_POS.shape}), but players shape is {players.shape}"


def simulate_passes(
    # Input data
    PLAYER_POS,  # F x P x 4[x, y, vx, vy], player positions
    BALL_POS,  # F x 2[x, y], ball positions
    phi_grid,  # F x PHI, pass angles
    v0_grid,  # F x V0, pass speeds
    passer_teams,  # F, frame-wise team of passers
    player_teams,  # P, player teams
    players=None,  # P, players
    passers=None,  # F, frame-wise passer
    exclude_passer=False,
    playing_direction=None,
    respect_offside=False,
    fields_to_return=ALL_OPTIONAL_FIELDS_TO_RETURN,
    x_pitch_min=-52.5, x_pitch_max=52.5, y_pitch_min=-34, y_pitch_max=34,

    # Model parameters
    pass_start_location_offset=_DEFAULT_PASS_START_LOCATION_OFFSET,
    time_offset_ball=_DEFAULT_TIME_OFFSET_BALL,
    radial_gridsize=_DEFAULT_RADIAL_GRIDSIZE,
    b0=_DEFAULT_B0,
    b1=_DEFAULT_B1,
    player_velocity=_DEFAULT_PLAYER_VELOCITY,
    keep_inertial_velocity=_DEFAULT_KEEP_INERTIAL_VELOCITY,
    use_max=_DEFAULT_USE_MAX,
    v_max=_DEFAULT_V_MAX,
    a_max=_DEFAULT_A_MAX,
    inertial_seconds=_DEFAULT_INERTIAL_SECONDS,
    tol_distance=_DEFAULT_TOL_DISTANCE,
    use_approx_two_point=_DEFAULT_USE_APPROX_TWO_POINT,
    v0_prob_aggregation_mode=_DEFAULT_V0_PROB_AGGREGATION_MODE,
    normalize=_DEFAULT_NORMALIZE,
    use_efficient_sigmoid=_DEFAULT_USE_EFFICIENT_SIGMOID,

    factor=_DEFAULT_FACTOR,
    factor2=_DEFAULT_FACTOR2,
) -> SimulationResult:
    """ Calculate the pass simulation model using numpy matrices - Core functionality of this package

    # Simulate a pass from player A straight to the right towards a defender B who is 50m away.
    >>> res = simulate_passes(np.array([[[0, 0, 0, 0], [50, 0, 0, 0]]]), np.array([[0, 0]]), np.array([[0]]), np.array([[10]]), np.array([0]), np.array([0, 1]), players=np.array(["A", "B"]), passers=np.array(["A"]), radial_gridsize=15)
    >>> res.defense_poss_density.shape, res.defense_poss_density
    ((1, 1, 13), array([[[8.89786731e-05, 1.64897230e-04, 1.29669104e-03, 1.94280126e-01,
             1.91894891e-01, 1.88063562e-01, 1.77544360e-01, 3.89546322e-02,
             2.01864630e-03, 1.02515215e-04, 5.17596953e-06, 2.60590596e-07,
             1.30974323e-08]]]))
    >>> res.attack_cum_prob.shape, res.attack_cum_prob  # F x PHI x T
    ((1, 1, 13), array([[[0.        , 0.01097577, 0.01807728, 0.02306136, 0.02424135,
             0.02432508, 0.02433268, 0.02433416, 0.02433455, 0.02433456,
             0.02433456, 0.02433456, 0.02433456]]]))
    >>> res.phi_grid.shape, res.phi_grid
    ((1, 1), array([[0]]))
    >>> res.r_grid.shape, res.r_grid
    ((13,), array([ -1.52453403,  13.47546597,  28.47546597,  43.47546597,
            58.47546597,  73.47546597,  88.47546597, 103.47546597,
           118.47546597, 133.47546597, 148.47546597, 163.47546597,
           178.47546597]))
    >>> res.x_grid.shape, res.x_grid
    ((1, 1, 13), array([[[ -1.52453403,  13.47546597,  28.47546597,  43.47546597,
              58.47546597,  73.47546597,  88.47546597, 103.47546597,
             118.47546597, 133.47546597, 148.47546597, 163.47546597,
             178.47546597]]]))
    >>> res.y_grid.shape, res.y_grid
    ((1, 1, 13), array([[[0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]]]))
    """
    if not set(fields_to_return) <= set(ALL_OPTIONAL_FIELDS_TO_RETURN):
        invalid_fields = set(fields_to_return) - set(ALL_OPTIONAL_FIELDS_TO_RETURN)
        raise ValueError(f"fields_to_return contains unknown fields: {invalid_fields}. Available: {ALL_OPTIONAL_FIELDS_TO_RETURN}")

    _assert_matrix_consistency(PLAYER_POS, BALL_POS, phi_grid, v0_grid, passer_teams, player_teams, players, passers)

    def _should_return_any_of(fields):
        return any([field in fields for field in fields_to_return])

    # Pre-processing: Treat offside players like air
    if respect_offside:
        if playing_direction is None:
            raise ValueError("'playing_direction' must be provided if 'respect_offside' is True")

        PLAYERS_NORM_X = PLAYER_POS[:, :, 0] * playing_direction[:, np.newaxis]  # F x P
        BALL_NORM_X = BALL_POS[:, 0] * playing_direction

        is_attacking_team = passer_teams[:, np.newaxis] == player_teams[np.newaxis, :]  # F x P
        SECOND_LAST_DEFENDER_NORM_X = np.ma.sort(np.ma.array(PLAYERS_NORM_X, mask=is_attacking_team), axis=1, endwith=False)[::-1][:, -2]
        X_OFFSIDE_LINE = np.maximum(SECOND_LAST_DEFENDER_NORM_X, BALL_NORM_X)  # F

        PLAYER_IS_OFFSIDE = is_attacking_team & (PLAYERS_NORM_X > X_OFFSIDE_LINE[:, np.newaxis]) & (PLAYERS_NORM_X > 0)

        if passers is not None:
            player_is_passer = passers[:, np.newaxis] == players[np.newaxis, :]  # F x P
            PLAYER_IS_OFFSIDE = PLAYER_IS_OFFSIDE & (~player_is_passer)  # F x P

        try:
            PLAYER_POS[PLAYER_IS_OFFSIDE, :] = np.nan
        except ValueError:
            warnings.warn("Offside not properly detectable, maybe too few defenders. Ignoring offside.")

    ### 1. Calculate ball trajectory
    # 1.1 Calculate spatial grid
    max_pass_length = np.sqrt((x_pitch_max - x_pitch_min) ** 2 + (y_pitch_max - y_pitch_min) ** 2) + radial_gridsize * 3  # T
    D_BALL_SIM = np.arange(pass_start_location_offset, max_pass_length + pass_start_location_offset + radial_gridsize, radial_gridsize)  # T

    # 1.2 Calculate temporal grid
    T_BALL_SIM = constant_velocity_time_to_arrive_1d(
        x=D_BALL_SIM[0], v=v0_grid[:, :, np.newaxis], x_target=D_BALL_SIM[np.newaxis, np.newaxis, :],
    )  # F x V0 x T
    T_BALL_SIM += time_offset_ball

    # 1.3 Calculate 2D points along ball trajectory
    cos_phi, sin_phi = np.cos(phi_grid), np.sin(phi_grid)  # F x PHI
    X_BALL_SIM = BALL_POS[:, 0][:, np.newaxis, np.newaxis] + cos_phi[:, :, np.newaxis] * D_BALL_SIM[np.newaxis, np.newaxis, :]  # F x PHI x T
    Y_BALL_SIM = BALL_POS[:, 1][:, np.newaxis, np.newaxis] + sin_phi[:, :, np.newaxis] * D_BALL_SIM[np.newaxis, np.newaxis, :]  # F x PHI x T

    ### 2 Calculate player interception rates
    # 2.1 Calculate time to arrive for each player along ball trajectory
    if use_approx_two_point:
        TTA_PLAYERS = approx_two_point_time_to_arrive(  # F x P x PHI x T
            x=PLAYER_POS[:, :, 0][:, :, np.newaxis, np.newaxis], y=PLAYER_POS[:, :, 1][:, :, np.newaxis, np.newaxis],
            vx=PLAYER_POS[:, :, 2][:, :, np.newaxis, np.newaxis], vy=PLAYER_POS[:, :, 3][:, :, np.newaxis, np.newaxis],
            x_target=X_BALL_SIM[:, np.newaxis, :, :], y_target=Y_BALL_SIM[:, np.newaxis, :, :],

            # Parameters
            use_max=use_max, velocity=player_velocity, keep_inertial_velocity=keep_inertial_velocity, v_max=v_max,
            a_max=a_max, inertial_seconds=inertial_seconds, tol_distance=tol_distance,
        )
    else:
        TTA_PLAYERS = constant_velocity_time_to_arrive(  # F x P x PHI x T
            x=PLAYER_POS[:, :, 0][:, :, np.newaxis, np.newaxis], y=PLAYER_POS[:, :, 1][:, :, np.newaxis, np.newaxis],
            x_target=X_BALL_SIM[:, np.newaxis, :, :], y_target=Y_BALL_SIM[:, np.newaxis, :, :],

            # Parameter
            player_velocity=player_velocity,
        )

    if exclude_passer:
        if passers is None:
            raise ValueError("'passers' must be provided if 'exclude_passer' is True")
        i_passers_to_exclude = np.array([list(players).index(passer) for passer in passers])
        i_frames = np.arange(TTA_PLAYERS.shape[0])
        TTA_PLAYERS[i_frames, i_passers_to_exclude, :, :] = np.inf  # F x P x PHI x T

    TTA_PLAYERS = np.nan_to_num(TTA_PLAYERS, nan=np.inf)  # Handle players not participating in the game by setting their TTA to infinity
    gc.collect()

    # 2.2 Transform time to arrive into interception rates
    TMP = TTA_PLAYERS[:, :, np.newaxis, :, :] - T_BALL_SIM[:, np.newaxis, :, np.newaxis, :]  # F x P x PHI x T - F x PHI x T = F x P x V0 x PHI x T
    with np.errstate(over='ignore'):  # overflow leads to inf which will be handled gracefully later
        TMP[:] = b0 + b1 * TMP  # 1 + 1 * F x P x V0 x PHI x T = F x P x V0 x PHI x T
    with np.errstate(invalid='ignore'):  # inf -> nan
        TMP[:] = _approximate_sigmoid(TMP) if use_efficient_sigmoid else _sigmoid(TMP)
    TMP = np.nan_to_num(TMP, nan=0)  # F x P x V0 x PHI x T, gracefully handle overflow
    DT = T_BALL_SIM[:, :, 1] - T_BALL_SIM[:, :, 0]  # F x V0
    ar_time = TMP / (factor * (v0_grid ** (-factor2))[:, np.newaxis, :, np.newaxis, np.newaxis])  # F x P x V0 x PHI x T
    del TMP
    gc.collect()

    ## 3. Use interception rates to calculate probabilities
    # 3.1 Sums of interception rates over players
    sum_ar = np.nansum(ar_time, axis=1) if _should_return_any_of(["attack_prob_density", "defense_prob_density", "player_prob_density", "attack_cum_prob", "defense_cum_prob", "player_cum_prob", "p0_density", "cum_p0"]) else None  # F x V0 x PHI x T
    player_is_attacking = (player_teams[np.newaxis, :] == passer_teams[:, np.newaxis]) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density", "attack_cum_prob", "defense_cum_prob", "attack_prob_density", "defense_prob_density"]) else None  # F x P
    sum_ar_att = np.nansum(np.where(player_is_attacking[:, :, np.newaxis, np.newaxis, np.newaxis], ar_time, 0), axis=1) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x V0 x PHI x T
    sum_ar_def = np.nansum(np.where(~player_is_attacking[:, :, np.newaxis, np.newaxis, np.newaxis], ar_time, 0), axis=1) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x V0 x PHI x T

    # 3.2 Integral of sum of interception rates
    int_sum_ar = scipy.integrate.cumulative_trapezoid(y=sum_ar, x=T_BALL_SIM[:, :, np.newaxis, :], initial=0, axis=-1) if _should_return_any_of(["attack_prob_density", "defense_prob_density", "player_prob_density", "attack_cum_prob", "defense_cum_prob", "player_cum_prob", "p0_density", "cum_p0"]) else None  # F x V0 x PHI x T
    int_sum_ar_att = scipy.integrate.cumulative_trapezoid(y=sum_ar_att, x=T_BALL_SIM[:, :, np.newaxis, :], initial=0, axis=-1) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x V0 x PHI x T
    int_sum_ar_def = scipy.integrate.cumulative_trapezoid(y=sum_ar_def, x=T_BALL_SIM[:, :, np.newaxis, :], initial=0, axis=-1) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x V0 x PHI x T

    # 3.3 Cumulative probability P0 from integrals
    cum_p0 = np.exp(-int_sum_ar) if _should_return_any_of(["attack_prob_density", "defense_prob_density", "player_prob_density", "attack_cum_prob", "defense_cum_prob", "player_cum_prob", "p0_density", "cum_p0"]) else None  # F x V0 x PHI x T, cumulative probability that no one intercepted
    cum_p0_only_att = np.exp(-int_sum_ar_att) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x V0 x PHI x T
    cum_p0_only_def = np.exp(-int_sum_ar_def) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x V0 x PHI x T
    cum_p0_only_opp = np.where(
        player_is_attacking[:, :, np.newaxis, np.newaxis, np.newaxis],
        cum_p0_only_def[:, np.newaxis, :, :, :], cum_p0_only_att[:, np.newaxis, :, :, :]
    ) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x P x V0 x PHI x T

    # 3.4 Probability density from P0
    dpr_over_dt = cum_p0[:, np.newaxis, :, :, :] * ar_time if _should_return_any_of(["attack_prob_density", "defense_prob_density", "player_prob_density", "attack_cum_prob", "defense_cum_prob", "player_cum_prob", "cum_p0"]) else None  # if "prob" in ptypes else None  # F x P x V0 x PHI x T
    dp0_over_dt = -cum_p0 * sum_ar if _should_return_any_of(["p0_density"]) else None  # F x V0 x PHI x T
    dpr_poss_over_dt = cum_p0_only_opp * ar_time if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # if "poss" in ptypes else None  # F x P x V0 x PHI x T

    # 3.5 Cumulative probability
    pr_cum_prob = scipy.integrate.cumulative_trapezoid(  # F x P x V0 x PHI x T, cumulative probability that player P intercepted
        y=dpr_over_dt,  # F x P x V0 x PHI x T
        x=T_BALL_SIM[:, np.newaxis, :, np.newaxis, :],  # F x V0 x T
        initial=0, axis=-1,
    ) if _should_return_any_of(["attack_cum_prob", "defense_cum_prob", "player_cum_prob", "cum_p0"]) else None

    # 3.6. Go from dt -> dx
    DX = radial_gridsize  # ok because we use an equally spaced grid
    dpr_over_dx = dpr_over_dt * DT[:, np.newaxis, :, np.newaxis, np.newaxis] / DX if _should_return_any_of(["attack_prob_density", "defense_prob_density", "player_prob_density"]) else None  # F x P x V0 x PHI x T
    del dpr_over_dt
    gc.collect()
    dp0_over_dx = dp0_over_dt * DT[:, :, np.newaxis, np.newaxis] / DX if _should_return_any_of(["p0_density"]) else None  # F x V0 x PHI x T
    del dp0_over_dt
    gc.collect()
    dpr_poss_over_dx = dpr_poss_over_dt * DT[:, np.newaxis, :, np.newaxis, np.newaxis] / DX if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x P x V0 x PHI x T
    del dpr_poss_over_dt
    gc.collect()

    # 3.7 Aggregate over v0
    player_prob_density = (np.mean(dpr_over_dx, axis=2) if v0_prob_aggregation_mode == "mean" else np.max(dpr_over_dx, axis=2)) if _should_return_any_of(["attack_prob_density", "defense_prob_density", "player_prob_density"]) else None  # F x P x PHI x T, Take the average over all V0 in v0_grid
    p0_density = (np.mean(dp0_over_dx, axis=1) if v0_prob_aggregation_mode == "mean" else np.min(dp0_over_dx, axis=1)) if _should_return_any_of(["p0_density"]) else None  # F x PHI x T
    player_poss_density = np.max(dpr_poss_over_dx, axis=2) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x P x PHI x T, np.max not supported yet with numba using axis https://github.com/numba/numba/issues/1269

    cum_p0_vagg = (np.mean(cum_p0, axis=1) if v0_prob_aggregation_mode == "mean" else np.min(cum_p0, axis=1)) if _should_return_any_of(["cum_p0"]) or normalize and _should_return_any_of(["attack_cum_prob", "defense_cum_prob", "player_cum_prob"]) else None  # F x PHI x T
    pr_cum_prob_vagg = (np.mean(pr_cum_prob, axis=2) if v0_prob_aggregation_mode == "mean" else np.max(pr_cum_prob, axis=2)) if _should_return_any_of(["attack_cum_prob", "defense_cum_prob", "player_cum_prob", "cum_p0"]) else None  # F x P x PHI x T

    # 3.8 Normalize
    if normalize:
        # Normalize cumulative probability
        p_cum_sum = cum_p0_vagg + pr_cum_prob_vagg.sum(axis=1) if _should_return_any_of(["attack_cum_prob", "defense_cum_prob", "player_cum_prob", "cum_p0"]) else None  # F x PHI x T
        cum_p0_vagg = cum_p0_vagg / p_cum_sum if _should_return_any_of(["cum_p0"]) else None
        pr_cum_prob_vagg = pr_cum_prob_vagg / p_cum_sum[:, np.newaxis, :, :] if _should_return_any_of(["attack_cum_prob", "defense_cum_prob", "player_cum_prob"]) else None  # F x P x PHI x T

        # Normalize possibility density
        dpr_over_dx_vagg_poss_times_dx = player_poss_density * DX if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x P x PHI x T
        num_max = np.max(dpr_over_dx_vagg_poss_times_dx, axis=(1, 3)) if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x PHI
        with np.errstate(invalid='ignore'):
            player_poss_density = player_poss_density / num_max[:, np.newaxis, :, np.newaxis] if _should_return_any_of(["attack_cum_poss", "attack_poss_density", "defense_cum_poss", "defense_poss_density", "player_cum_poss", "player_poss_density"]) else None  # F x P x PHI x T

        # TODO: Normalization is hard because the prob-/possibilities are time-dependent AND need to be normalized w.r.t both the player- and the time-axis.

    # 3.9 Aggregate over players (Individual level -> Team level)
    attack_prob_density = np.nansum(np.where(player_is_attacking[:, :, np.newaxis, np.newaxis], player_prob_density, 0), axis=1) if _should_return_any_of(["attack_prob_density"]) else None  # F x PHI x T
    defense_prob_density = np.nansum(np.where(~player_is_attacking[:, :, np.newaxis, np.newaxis], player_prob_density, 0), axis=1) if _should_return_any_of(["defense_prob_density"]) else None  # F x PHI x T
    attack_poss_density = np.nanmax(np.where(player_is_attacking[:, :, np.newaxis, np.newaxis], player_poss_density, 0), axis=1) if _should_return_any_of(["attack_cum_poss", "attack_poss_density"]) else None  # F x PHI x T
    defense_poss_density = np.nanmax(np.where(~player_is_attacking[:, :, np.newaxis, np.newaxis], player_poss_density, 0), axis=1) if _should_return_any_of(["defense_cum_poss", "defense_poss_density"]) else None  # F x PHI x T
    attack_cum_prob = np.nansum(np.where(player_is_attacking[:, :, np.newaxis, np.newaxis], pr_cum_prob_vagg, 0), axis=1) if _should_return_any_of(["attack_cum_prob"]) else None  # F x PHI x T
    defense_cum_prob = np.nansum(np.where(~player_is_attacking[:, :, np.newaxis, np.newaxis], pr_cum_prob_vagg, 0), axis=1) if _should_return_any_of(["defense_cum_prob"]) else None  # F x PHI x T

    player_cum_poss = np.maximum.accumulate(player_poss_density, axis=-1) * radial_gridsize if _should_return_any_of(["player_cum_poss"]) else None  # TODO cleaner would be to move this earlier?
    attack_cum_poss = np.maximum.accumulate(attack_poss_density, axis=-1) * radial_gridsize if _should_return_any_of(["attack_cum_poss"]) else None  # possibility CDF uses cummax instead of cumsum to emerge from PDF
    defense_cum_poss = np.maximum.accumulate(defense_poss_density, axis=-1) * radial_gridsize if _should_return_any_of(["defense_cum_poss"]) else None

    result = SimulationResult(
        # Team-level prob-/possibilities (cumulative and densities) along simulated ball trajectories
        attack_cum_prob=attack_cum_prob,  # F x PHI x T
        attack_cum_poss=attack_cum_poss,  # F x PHI x T
        attack_prob_density=attack_prob_density,  # F x PHI x T
        attack_poss_density=attack_poss_density,  # F x PHI x T
        defense_cum_prob=defense_cum_prob,  # F x PHI x T
        defense_cum_poss=defense_cum_poss,  # F x PHI x T
        defense_prob_density=defense_prob_density,  # F x PHI x T
        defense_poss_density=defense_poss_density,  # F x PHI x T

        # Player-specific prob-/possibilities
        player_cum_prob=pr_cum_prob_vagg,  # F x P x PHI x T
        player_cum_poss=player_cum_poss,  # F x P x PHI x T
        player_prob_density=player_prob_density,  # F x P x PHI x T
        player_poss_density=player_poss_density,  # F x P x PHI x T

        # Complementary proability
        cum_p0=cum_p0_vagg,  # F x PHI x T
        p0_density=p0_density,  # F x PHI x T

        # Trajectory grids
        phi_grid=phi_grid,  # F x PHI
        r_grid=D_BALL_SIM,  # T
        x_grid=X_BALL_SIM,  # F x PHI x T
        y_grid=Y_BALL_SIM,  # F x PHI x T
    )

    # Set fields not to return to zero
    for field in _result_fields:
        if field in ["phi_grid", "r_grid", "x_grid", "y_grid"]:
            continue
        if field not in fields_to_return:
            result = result._replace(**{field: None})

    return result


def simulate_passes_chunked(
    # Input data
    PLAYER_POS,
    BALL_POS,
    phi_grid,
    v0_grid,
    passer_teams,
    player_teams,
    players=None,
    passers=None,  # F, frame-wise passer
    exclude_passer=False,
    playing_direction=None,
    respect_offside=False,
    x_pitch_min=-52.5, x_pitch_max=52.5, y_pitch_min=-34, y_pitch_max=34,

    # Options
    use_progress_bar=True,
    chunk_size=150,
    fields_to_return=ALL_OPTIONAL_FIELDS_TO_RETURN,

    # Model parameters
    pass_start_location_offset=_DEFAULT_PASS_START_LOCATION_OFFSET,
    time_offset_ball=_DEFAULT_TIME_OFFSET_BALL,
    radial_gridsize=_DEFAULT_RADIAL_GRIDSIZE,
    b0=_DEFAULT_B0,
    b1=_DEFAULT_B1,
    player_velocity=_DEFAULT_PLAYER_VELOCITY,
    keep_inertial_velocity=_DEFAULT_KEEP_INERTIAL_VELOCITY,
    use_max=_DEFAULT_USE_MAX,
    v_max=_DEFAULT_V_MAX,
    a_max=_DEFAULT_A_MAX,
    inertial_seconds=_DEFAULT_INERTIAL_SECONDS,
    tol_distance=_DEFAULT_TOL_DISTANCE,
    use_approx_two_point=_DEFAULT_USE_APPROX_TWO_POINT,
    v0_prob_aggregation_mode=_DEFAULT_V0_PROB_AGGREGATION_MODE,
    normalize=_DEFAULT_NORMALIZE,
    use_efficient_sigmoid=_DEFAULT_USE_EFFICIENT_SIGMOID,

    factor=_DEFAULT_FACTOR,
    factor2=_DEFAULT_FACTOR2,
) -> SimulationResult:
    """
    Execute pass simulation in chunks to avoid OOM.

    >>> res = simulate_passes_chunked(np.array([[[0, 0, 0, 0], [50, 0, 0, 0]]]), np.array([[0, 0]]), np.array([[0]]), np.array([[10]]), np.array([0]), np.array([0, 1]), players=np.array(["A", "B"]), passers=np.array(["A"]), radial_gridsize=15)
    >>> res.defense_poss_density.shape, res.defense_poss_density
    ((1, 1, 13), array([[[8.89786731e-05, 1.64897230e-04, 1.29669104e-03, 1.94280126e-01,
             1.91894891e-01, 1.88063562e-01, 1.77544360e-01, 3.89546322e-02,
             2.01864630e-03, 1.02515215e-04, 5.17596953e-06, 2.60590596e-07,
             1.30974323e-08]]]))
    """
    if chunk_size is None or chunk_size <= 0:
        chunk_size = PLAYER_POS.shape[0]
    _assert_matrix_consistency(PLAYER_POS, BALL_POS, phi_grid, v0_grid, passer_teams, player_teams, players, passers)

    F = PLAYER_POS.shape[0]

    i_chunks = list(range(0, F, chunk_size))

    full_result = None

    if use_progress_bar:
        i_chunks = progress_bar(i_chunks, desc="Simulating passes", total=len(i_chunks), unit="chunk")

    for chunk_nr, i in enumerate(i_chunks):
        i_chunk_end = min(i + chunk_size, F)

        PLAYER_POS_chunk = PLAYER_POS[i:i_chunk_end, ...]
        BALL_POS_chunk = BALL_POS[i:i_chunk_end, ...]
        phi_grid_chunk = phi_grid[i:i_chunk_end, ...]
        v0_grid_chunk = v0_grid[i:i_chunk_end, ...]
        passer_team_chunk = passer_teams[i:i_chunk_end, ...]
        if passers is not None:
            passers_chunk = passers[i:i_chunk_end, ...]
        else:
            passers_chunk = None
        if playing_direction is not None:
            playing_direction_chunk = playing_direction[i:i_chunk_end, ...]
        else:
            playing_direction_chunk = None

        result = simulate_passes(
            PLAYER_POS_chunk, BALL_POS_chunk, phi_grid_chunk, v0_grid_chunk, passer_team_chunk, player_teams, players,
            passers_chunk,
            exclude_passer,
            playing_direction_chunk,
            respect_offside,
            fields_to_return,
            x_pitch_min, x_pitch_max, y_pitch_min, y_pitch_max,
            pass_start_location_offset,
            time_offset_ball,
            radial_gridsize,
            b0,
            b1,
            player_velocity,
            keep_inertial_velocity,
            use_max,
            v_max,
            a_max,
            inertial_seconds,
            tol_distance,
            use_approx_two_point,
            v0_prob_aggregation_mode,
            normalize,
            use_efficient_sigmoid,
            factor,
            factor2,
        )

        if full_result is None:
            full_result = result
        else:
            full_p_cum = np.concatenate([full_result.attack_cum_prob, result.attack_cum_prob], axis=0) if full_result.attack_cum_prob is not None else None
            full_poss_cum = np.concatenate([full_result.attack_cum_poss, result.attack_cum_poss], axis=0) if full_result.attack_cum_poss is not None else None
            full_p_density = np.concatenate([full_result.attack_poss_density, result.attack_poss_density], axis=0) if full_result.attack_poss_density is not None else None
            full_prob_density = np.concatenate([full_result.attack_prob_density, result.attack_prob_density], axis=0) if full_result.attack_prob_density is not None else None
            full_p_cum_def = np.concatenate([full_result.defense_cum_prob, result.defense_cum_prob], axis=0) if full_result.defense_cum_prob is not None else None
            full_defense_cum_poss = np.concatenate([full_result.defense_cum_poss, result.defense_cum_poss], axis=0) if full_result.defense_cum_poss is not None else None
            full_p_density_def = np.concatenate([full_result.defense_poss_density, result.defense_poss_density], axis=0) if full_result.defense_poss_density is not None else None
            full_defense_prob_density = np.concatenate([full_result.defense_prob_density, result.defense_prob_density], axis=0) if full_result.defense_prob_density is not None else None
            full_cum_p0 = np.concatenate([full_result.cum_p0, result.cum_p0], axis=0) if full_result.cum_p0 is not None else None
            full_p0_density = np.concatenate([full_result.p0_density, result.p0_density], axis=0) if full_result.p0_density is not None else None
            full_phi = np.concatenate([full_result.phi_grid, result.phi_grid], axis=0) if full_result.phi_grid is not None else None
            full_x0 = np.concatenate([full_result.x_grid, result.x_grid], axis=0) if full_result.x_grid is not None else None
            full_y0 = np.concatenate([full_result.y_grid, result.y_grid], axis=0) if full_result.y_grid is not None else None
            full_player_prob_density = np.concatenate([full_result.player_prob_density, result.player_prob_density], axis=0) if full_result.player_prob_density is not None else None
            full_player_poss_density = np.concatenate([full_result.player_poss_density, result.player_poss_density], axis=0) if full_result.player_poss_density is not None else None
            full_player_cum_prob = np.concatenate([full_result.player_cum_prob, result.player_cum_prob], axis=0) if full_result.player_cum_prob is not None else None
            full_player_cum_poss = np.concatenate([full_result.player_cum_poss, result.player_cum_poss], axis=0) if full_result.player_cum_poss is not None else None
            full_result = SimulationResult(
                attack_cum_poss=full_poss_cum,
                attack_cum_prob=full_p_cum,
                attack_poss_density=full_p_density,
                attack_prob_density=full_prob_density,
                defense_cum_poss=full_defense_cum_poss,
                defense_cum_prob=full_p_cum_def,
                defense_poss_density=full_p_density_def,
                defense_prob_density=full_defense_prob_density,
                cum_p0=full_cum_p0,
                p0_density=full_p0_density,
                player_prob_density=full_player_prob_density,
                player_poss_density=full_player_poss_density,
                player_cum_prob=full_player_cum_prob,
                player_cum_poss=full_player_cum_poss,
                phi_grid=full_phi,
                r_grid=full_result.r_grid,
                x_grid=full_x0,
                y_grid=full_y0,
            )

    return full_result


def clip_simulation_result_to_pitch(
    simulation_result: SimulationResult, x_pitch_min=-52.5, x_pitch_max=52.5, y_pitch_min=-34, y_pitch_max=34,
) -> SimulationResult:
    """
    Set all data points that are outside the pitch to zero (e.g. for DAS computation)

    >>> res = simulate_passes(np.array([[[0, 0, 0, 0], [50, 0, 0, 0]]]), np.array([[0, 0]]), np.array([[0]]), np.array([[10]]), np.array([0]), np.array([0, 1]), players=np.array(["A", "B"]), passers=np.array(["A"]), radial_gridsize=15)
    >>> res.defense_poss_density
    array([[[8.89786731e-05, 1.64897230e-04, 1.29669104e-03, 1.94280126e-01,
             1.91894891e-01, 1.88063562e-01, 1.77544360e-01, 3.89546322e-02,
             2.01864630e-03, 1.02515215e-04, 5.17596953e-06, 2.60590596e-07,
             1.30974323e-08]]])
    >>> clip_simulation_result_to_pitch(res).defense_poss_density
    array([[[8.89786731e-05, 1.64897230e-04, 1.29669104e-03, 1.94280126e-01,
             0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
             0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
             0.00000000e+00]]])
    """
    x = simulation_result.x_grid
    y = simulation_result.y_grid

    if x_pitch_min > x_pitch_max:
        raise ValueError("x_pitch_min must be smaller than x_pitch_max")
    if y_pitch_min > y_pitch_max:
        raise ValueError("y_pitch_min must be smaller than y_pitch_max")

    on_pitch_mask = ((x >= x_pitch_min) & (x <= x_pitch_max) & (y >= y_pitch_min) & (y <= y_pitch_max))  # F x PHI x T

    for cumulative_field in ["attack_cum_prob", "attack_cum_poss", "defense_cum_prob", "defense_cum_poss", "cum_p0", "player_cum_prob", "player_cum_poss"]:
        cum_field_data = getattr(simulation_result, cumulative_field)
        if cum_field_data is None:
            continue

        on_pitch_mask_field = on_pitch_mask if len(cum_field_data.shape) == 3 else np.repeat(on_pitch_mask[:, np.newaxis, :, :], cum_field_data.shape[1], axis=1)

        start_value = 0 if cumulative_field != "cum_p0" else 1
        updated_cum = cum_field_data.copy()
        for i in range(cum_field_data.shape[-1]):
            updated_cum[..., i] = np.where(on_pitch_mask_field[..., i], cum_field_data[..., i], updated_cum[..., i-1] if i > 0 else start_value)
            if cumulative_field == "attack_cum_prob":
                if i > 0:
                    assert np.all(on_pitch_mask_field[..., i] | (updated_cum[..., i] == updated_cum[..., i-1]))

        simulation_result = simulation_result._replace(**{cumulative_field: updated_cum})

    density_replacement_value = 0
    simulation_result = simulation_result._replace(
        attack_prob_density=np.where(on_pitch_mask, simulation_result.attack_prob_density, density_replacement_value) if simulation_result.attack_prob_density is not None else None,
        attack_poss_density=np.where(on_pitch_mask, simulation_result.attack_poss_density, density_replacement_value) if simulation_result.attack_poss_density is not None else None,
        defense_prob_density=np.where(on_pitch_mask, simulation_result.defense_prob_density, density_replacement_value) if simulation_result.defense_prob_density is not None else None,
        defense_poss_density=np.where(on_pitch_mask, simulation_result.defense_poss_density, density_replacement_value) if simulation_result.defense_poss_density is not None else None,
        p0_density=np.where(on_pitch_mask, simulation_result.p0_density, density_replacement_value) if simulation_result.p0_density is not None else None,
        player_prob_density=np.where(on_pitch_mask[:, np.newaxis, :, :], simulation_result.player_prob_density, density_replacement_value) if simulation_result.player_prob_density is not None else None,
        player_poss_density=np.where(on_pitch_mask[:, np.newaxis, :, :], simulation_result.player_poss_density, density_replacement_value) if simulation_result.player_poss_density is not None else None,
    )
    return simulation_result


def integrate_surfaces(result: SimulationResult, x_pitch_min=-52.5, x_pitch_max=52.5, y_pitch_min=-34, y_pitch_max=34):
    """
    Integrate attacking possibility density in result to obtain surface area (AS/DAS)

    >>> res = simulate_passes(np.array([[[0, 0, 0, 0], [50, 0, 0, 0]]]), np.array([[0, 0]]), np.array([[0, 1*np.pi/3, 2*np.pi/3]]), np.array([[10, 10, 10]]), np.array([0]), np.array([0, 1]), radial_gridsize=15)
    >>> res.attack_poss_density
    array([[[1.03407896e-03, 4.34113430e-04, 5.27060765e-04, 1.51129699e-04,
             1.07173438e-05, 9.21090837e-07, 1.63818876e-07, 2.68226445e-07,
             1.35596313e-08, 6.79782508e-10, 3.40267229e-11, 1.70221067e-12,
             8.51279969e-14],
            [1.03407896e-03, 4.34208810e-04, 5.31611162e-04, 6.84624667e-04,
             9.50183615e-04, 3.61836440e-04, 6.51260823e-05, 1.07221499e-04,
             5.43985030e-06, 2.73437156e-07, 1.37154339e-08, 6.87289329e-10,
             3.44205118e-11],
            [1.03407896e-03, 4.34301416e-04, 5.32301463e-04, 6.87956371e-04,
             9.73593533e-04, 1.66948127e-03, 5.92149182e-03, 1.93318292e-01,
             1.94810292e-01, 1.94616189e-01, 1.94033526e-01, 1.93218568e-01,
             1.92180048e-01]]])
    >>> integrate_surfaces(res)
    Areas(attack_prob=array([10.74287034]), attack_poss=array([10.92307646]), defense_prob=array([224.5886554]), defense_poss=array([1002.23676372]), player_prob=array([[ 10.74287034, 224.5886554 ]]), player_poss=array([[  10.92307646, 1002.23676372]]))
    """
    result = clip_simulation_result_to_pitch(result, x_pitch_min, x_pitch_max, y_pitch_min, y_pitch_max)

    # 1. Get r-part of area elements
    r_grid = result.r_grid  # T

    r_lower_bounds = np.zeros_like(r_grid)  # Initialize with zeros
    r_lower_bounds[1:] = (r_grid[:-1] + r_grid[1:]) / 2  # Midpoint between current and previous element
    r_lower_bounds[0] = r_grid[0]  # Set lower bound for the first element

    r_upper_bounds = np.zeros_like(r_grid)  # Initialize with zeros
    r_upper_bounds[:-1] = (r_grid[:-1] + r_grid[1:]) / 2  # Midpoint between current and next element
    r_upper_bounds[-1] = r_grid[-1]  # Arbitrarily high upper bound for the last element

    dr = r_upper_bounds - r_lower_bounds  # T

    # 2. Get phi-part of area elements
    phi_grid = result.phi_grid  # F x PHI

    phi_lower_bounds = np.zeros_like(phi_grid)  # F x PHI
    phi_lower_bounds[:, 1:] = (phi_grid[:, :-1] + phi_grid[:, 1:]) / 2  # Midpoint between current and previous element
    phi_lower_bounds[:, 0] = phi_grid[:, 0]

    phi_upper_bounds = np.zeros_like(phi_grid)  # Initialize with zeros
    phi_upper_bounds[:, :-1] = (phi_grid[:, :-1] + phi_grid[:, 1:]) / 2  # Midpoint between current and next element
    phi_upper_bounds[:, -1] = phi_grid[:, -1]  # Arbitrarily high upper bound for the last element

    dphi = phi_upper_bounds - phi_lower_bounds  # F x PHI

    # 3. Calculate area elements
    outer_bound_circle_slice_area = dphi[:, :, np.newaxis]/(2*np.pi) * (np.pi * r_upper_bounds[np.newaxis, np.newaxis, :]**2)  # T
    inner_bound_circle_slice_area = dphi[:, :, np.newaxis]/(2*np.pi) * (np.pi * r_lower_bounds[np.newaxis, np.newaxis, :]**2)  # T
    dA = outer_bound_circle_slice_area - inner_bound_circle_slice_area  # F x PHI x T

    # 4. Calculate surface area
    Areas = collections.namedtuple("Areas", ["attack_prob", "attack_poss", "defense_prob", "defense_poss", "player_prob", "player_poss"])

    area_data = {}
    for attribute, team_field in [
        ("attack_prob", result.attack_prob_density),
        ("attack_poss", result.attack_poss_density),
        ("defense_prob", result.defense_prob_density),
        ("defense_poss", result.defense_poss_density),
    ]:
        if team_field is None:
            area_data[attribute] = None
        else:
            area_data[attribute] = np.sum(team_field * dr[np.newaxis, np.newaxis, :] * dA, axis=(1, 2))  # F x PHI x T

    for attribute, player_field in [
        ("player_prob", result.player_prob_density),
        ("player_poss", result.player_poss_density),
    ]:
        if player_field is None:
            area_data[attribute] = None
        else:
            probability_field = player_field * dr[np.newaxis, np.newaxis, np.newaxis, :] * dA[:, np.newaxis, :, :]
            area_data[attribute] = np.sum(probability_field, axis=(2, 3))  # F x P x PHI x T

    return Areas(**area_data)


def as_dangerous_result(result, danger, danger_weight):
    """
    Convert a simulation result to a dangerous simulation result by multiplying density with danger.

    >>> res = simulate_passes_chunked(np.array([[[0, 0, 0, 0], [50, 0, 0, 0]]]), np.array([[0, 0]]), np.array([[0]]), np.array([[10]]), np.array([0]), np.array([0, 1]), players=np.array(["A", "B"]), passers=np.array(["A"]), radial_gridsize=15)
    >>> res.defense_poss_density
    array([[[8.89786731e-05, 1.64897230e-04, 1.29669104e-03, 1.94280126e-01,
             1.91894891e-01, 1.88063562e-01, 1.77544360e-01, 3.89546322e-02,
             2.01864630e-03, 1.02515215e-04, 5.17596953e-06, 2.60590596e-07,
             1.30974323e-08]]])
    >>> danger = np.array([[[0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.0, 1.0]]])
    >>> dangerous_res = as_dangerous_result(res, danger, danger_weight=1)
    >>> dangerous_res.defense_poss_density
    array([[[0.00000000e+00, 1.64897230e-05, 2.59338208e-04, 5.82840377e-02,
             7.67579563e-02, 9.40317808e-02, 1.06526616e-01, 2.72682426e-02,
             1.61491704e-03, 9.22636939e-05, 5.17596953e-06, 2.60590596e-07,
             1.30974323e-08]]])
    """
    def weighted_multiplication(_danger, _space, weight=danger_weight):
        return (_danger ** (1 / weight)) * _space if weight is not None else _danger * _space

    return result._replace(
        attack_cum_poss=None,
        attack_cum_prob=None,
        attack_poss_density=weighted_multiplication(danger, result.attack_poss_density) if result.attack_poss_density is not None else None,
        attack_prob_density=weighted_multiplication(danger, result.attack_prob_density) if result.attack_prob_density is not None else None,
        defense_cum_poss=None,
        defense_cum_prob=None,
        defense_poss_density=weighted_multiplication(danger, result.defense_poss_density) if result.defense_poss_density is not None else None,
        defense_prob_density=weighted_multiplication(danger, result.defense_prob_density) if result.defense_prob_density is not None else None,
        player_cum_prob=None,
        player_cum_poss=None,
        player_poss_density=weighted_multiplication(danger[:, np.newaxis, :, :], result.player_poss_density) if result.player_poss_density is not None else None,
        player_prob_density=weighted_multiplication(danger[:, np.newaxis, :, :], result.player_prob_density) if result.player_prob_density is not None else None,
    )



import importlib

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest

import accessible_space


def _get_butterfly_data():
    df_tracking = pd.DataFrame({
        "frame_id": [0, 0, 0, 0],
        "player_id": ["a", "b", "x", "ball"],
        "x": [0, -50, 50, 0],
        "y": [0, 0, 0, 0],
        "vx": [0, 0, 0, 15],
        "vy": [0, 0, 0, 0],
        "team_id": ["H", "H", "A", "BALL"],
        "player_color": ["blue", "blue", "red", "black"],
        "team_in_possession": ["H"] * 4,
        "player_in_possession": ["a"] * 4,
        "attacking_direction": [1] * 4,
        # "frame_id": [0, 0, 0, 0, 1],
        # "player_id": ["a", "b", "x", "ball", "ball"],
        # "x": [0, -50, 50, 0, 1],
        # "y": [0, 0, 0, 0, 0],
        # "vx": [0, 0, 0, 15, 15],
        # "vy": [0, 0, 0, 0, 0],
        # "team_id": ["H", "H", "A", None, None],
        # "player_color": ["blue", "blue", "red", "black", "black"],
        # "team_in_possession": ["H"] * 5,
        # "player_in_possession": ["a"] * 5,
        # "attacking_direction": [1] * 5,
    })

    ### Plotting
    # plt.scatter(df_tracking["x"], df_tracking["y"], color=df_tracking["player_color"])
    # plt.show()

    df_pass_safe = pd.DataFrame({
        "frame_id": [0],
        "player_id": ["a"],
        "team_id": ["H"],
        "x": [0],
        "y": [0],
        "x_target": [-50],
        "y_target": [0],
        "v0": [15],
    })
    df_pass_risky = df_pass_safe.copy()
    df_pass_risky["x_target"] = 50

    return df_pass_safe, df_pass_risky, df_tracking

def _get_butterfly_data_with_nans():
    df_tracking = pd.DataFrame({
        "frame_id": [0, 0, 0, 0, 0, 0],
        "player_id": ["a", "b", "x", "ball", "c", "d"],
        "x": [0, -50, 50, 0, np.nan, np.nan],
        "y": [0, 0, 0, 0, np.nan, np.nan],
        "vx": [0, 0, 0, 15, np.nan, np.nan],
        "vy": [0, 0, 0, 0, np.nan, np.nan],
        "team_id": ["H", "H", "A", None, "H", "A"],
        "player_color": ["blue", "blue", "red", "black", "white", "white"],
        "team_in_possession": ["H"] * 6,
        "player_in_possession": ["a"] * 6,
        "attacking_direction": [1] * 6,
    })

    ### Plotting
    # plt.scatter(df_tracking["x"], df_tracking["y"], color=df_tracking["player_color"])
    # plt.show()

    df_pass_safe = pd.DataFrame({
        "frame_id": [0],
        "player_id": ["a"],
        "team_id": ["H"],
        "x": [0],
        "y": [0],
        "x_target": [-50],
        "y_target": [0],
        "v0": [15],
    })
    df_pass_risky = df_pass_safe.copy()
    df_pass_risky["x_target"] = 50

    return df_pass_safe, df_pass_risky, df_tracking


def _get_double_butterfly_data():
    df_tracking = pd.DataFrame({
        "frame_id": [0, 0, 0, 0, 0, 0],
        "player_id": ["a", "b", "c", "x", "y", "ball"],
        "team_id": ["H", "H", "H", "A", "A", None],
        "x": [0, -50, -10, 50, 10, 0],
        "y": [0, 0, 0, 0, 0, 0],
        "vx": [0, 0, 0, 0, 0, 0],
        "vy": [0, 0, 0, 0, 0, 0],
        "player_color": ["blue", "blue", "blue", "red", "red", "black"],
        "team_in_possession": ["H"] * 6,
        "player_in_possession": ["a"] * 6,
    })

    ### Plotting
    # plt.scatter(df_tracking["x"], df_tracking["y"], color=df_tracking["player_color"])
    # st.write(plt.gcf())
    # plt.show()

    return df_tracking


def _get_data_no_common_frames_in_event_and_tracking():
    df_tracking = pd.DataFrame({
        "frame_id": [0, 0, 0, 1, 1, 1, 2, 2, 2],
        "player_id": ["a", "b", "ball", "a", "b", "c", "a", "c", "ball"],
        "team_id": ["H", "A", None, "H", "A", "H", "H", "H", None],
        "x": [0, -50, -10, 50, 10, 0, 45, 55, 10],
        "y": [0, 0, 0, 0, 0, 0, 0, 0, 0],
        "vx": [0, 0, 0, 0, 0, 0, 0, 0, 0],
        "vy": [0, 0, 0, 0, 0, 0, 0, 0 ,0],
        "player_color": ["red", "blue", "black", "red", "blue", "red", "red", "red", "black"],
        "team_in_possession": ["H"] * 9,
        "player_in_possession": ["a"] * 9,
    })
    df_passes = pd.DataFrame({
        "frame_id": [0, 1, 2],
        "player_id": ["a", "a", "a"],
        "team_id": ["H", "H", "H"],
        "x": [0, 0, 0],
        "y": [0, 0, 0],
        "x_target": [-50, -50, -50],
        "y_target": [0, 0, 0],
        "v0": [15, 15, 15],
    })
    return df_passes, None, df_tracking


def test_offside():
    test_surface_plot(_get_butterfly_data)


def test_missing_ball_frames():  # it is not tested whether this results in valid results, it may even throw an exception!
    df_passes, _, df_tracking = _get_data_no_common_frames_in_event_and_tracking()
    with pytest.warns(UserWarning, match="Ball positions are missing"):
        try:
            accessible_space.get_expected_pass_completion(df_passes, df_tracking, respect_offside=False)
        except (ValueError, AssertionError):
            pass


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_non_matching_teams(_get_data):
    df_pass_safe, df_pass_risky, df_tracking = _get_data()
    df_pass_safe["team_id"] = "bömf"
    with pytest.raises(ValueError, match="do not match"):
        accessible_space.get_expected_pass_completion(df_pass_safe, df_tracking, respect_offside=False)


def test_no_poss_artifact_around_passer():
    _, _, df_tracking = _get_butterfly_data()

    # Using the default parameters should not lead to inaccessible artifact around passer
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, period_col=None, additional_fields_to_return=["attack_cum_poss"], respect_offside=False)

    ### Plotting
    # fig = accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "attack_poss_density", color="blue")
    # plt.xlim([-52.5, 52.5])
    # plt.ylim([-34, 34])
    assert (ret.simulation_result.attack_cum_poss[0, :, 1] > 0.5).any()


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_as_symmetry(_get_data):
    _, _, df_tracking = _get_data()

    for angles, v0_min, n_v0, gridsize in [
        (64, 0.01, 250, 1),
        (32, 0.01, 250, 1),
        (48, 0.01, 250, 1),
        (12, 0.01, 250, 1),
        (24, 0.01, 250, 1),
        (24, 0.001, 250, 1),
        (24, 0.0001, 250, 1),
        (24, 0.1, 250, 1),
        (24, 1, 250, 1),
        (24, 0.0001, 70, 1),
    ]:
        ret = accessible_space.get_dangerous_accessible_space(
            df_tracking, player_in_possession_col="player_in_possession", n_angles=angles, v0_min=v0_min,
            n_v0=n_v0, radial_gridsize=gridsize, period_col=None, respect_offside=False, exclude_passer=True,
        )

        ## Plotting
        # import streamlit as st
        # st.write(angles, v0_min, n_v0, gridsize, ret.acc_space.iloc[0])
        # accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "attack_poss_density", color="blue")
        # plt.xlim([-52.5, 52.5])
        # plt.ylim([-34, 34])
        # st.write(plt.gcf())
        # st.write("ret")
        # st.write("ret.acc_space", ret.acc_space)
        # st.write("ret.simulation_result.attack_poss_density[0]", ret.simulation_result.attack_poss_density[0])

        assert np.isclose(ret.acc_space.iloc[0], 3570, atol=1000)


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_xc_symmetry(_get_data):
    df_pass_safe, df_pass_risky, df_tracking = _get_data()

    ret_safe = accessible_space.get_expected_pass_completion(df_pass_safe, df_tracking, respect_offside=False)
    assert ret_safe.xc > 0.95
    ret_risky = accessible_space.get_expected_pass_completion(df_pass_risky, df_tracking, respect_offside=False)
    assert ret_risky.xc < 0.05

    # assert np.isclose(ret_safe.xc, 1 - ret_risky.xc, atol=1e-3)  # this is more a nice to have than a strict requirement at the moment (and it doesn't work with the current parameter settings)


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
@pytest.mark.parametrize("use_approx_two_point", [False, True])
@pytest.mark.parametrize("keep_inertial_velocity", [False, True])
@pytest.mark.parametrize("use_event_coordinates_as_ball_position", [False, True])
@pytest.mark.parametrize("use_fixed_v0", [False, True])
def test_xc_parameters(_get_data, use_approx_two_point, keep_inertial_velocity, use_event_coordinates_as_ball_position, use_fixed_v0):
    df_pass_safe, df_pass_risky, df_tracking = _get_data()

    ret_safe = accessible_space.get_expected_pass_completion(df_pass_safe, df_tracking, respect_offside=False, use_approx_two_point=use_approx_two_point, keep_inertial_velocity=keep_inertial_velocity, use_event_coordinates_as_ball_position=use_event_coordinates_as_ball_position, use_fixed_v0=use_fixed_v0, clip_to_pitch=False)
    assert ret_safe.xc[0] > 0.9

    df_tracking["vx"] = -df_tracking["vx"]
    df_tracking["x"] = -df_tracking["x"]
    ret_risky = accessible_space.get_expected_pass_completion(df_pass_safe, df_tracking, respect_offside=False, use_approx_two_point=use_approx_two_point, keep_inertial_velocity=keep_inertial_velocity, use_event_coordinates_as_ball_position=use_event_coordinates_as_ball_position, use_fixed_v0=use_fixed_v0, clip_to_pitch=False)
    df_tracking["vx"] = -df_tracking["vx"]
    df_tracking["x"] = -df_tracking["x"]

    assert ret_risky.xc[0] < 0.1
    # assert np.isclose(ret_safe.xc[0], 1 - ret_risky.xc[0], atol=1e-3)


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_coordinate_systems(_get_data):
    _, _, df_tracking = _get_data()
    for x_min, x_max, y_min, y_max in [
        (-52.5, 52.5, -34, 34),
        (0, 105, 0, 68),
        (-1, 1, -1, 1),
        (-2500, 2000, -2500, 1600),
        (-200, -20, -200, -20),
    ]:
        ret = accessible_space.get_dangerous_accessible_space(
            df_tracking, x_pitch_min=x_min, x_pitch_max=x_max, y_pitch_min=y_min, y_pitch_max=y_max, radial_gridsize=np.sqrt((x_max - x_min) ** 2 + (y_max - y_min) ** 2) / 50,
            period_col=None, respect_offside=False,
        )
        assert np.all(np.any(ret.simulation_result.x_grid <= x_min, axis=(1, 2)))
        assert np.all(np.any(ret.simulation_result.x_grid >= x_max, axis=(1, 2)))
        assert np.all(np.any(ret.simulation_result.y_grid <= y_min, axis=(1, 2)))
        assert np.all(np.any(ret.simulation_result.y_grid >= y_max, axis=(1, 2)))

        ### Plotting
        # plt.figure()
        # plt.xlim([x_min, x_max])
        # plt.ylim([y_min, y_max])
        # accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "attack_poss_density", color="blue")
        # accessible_space.plot_expected_completion_surface(ret.dangerous_result, 0, "attack_poss_density", color="red")
        # plt.title(f"Accessible space: {ret.acc_space.iloc[0]:.0f} m², DAS: {ret.das.iloc[0]:.2f} m²")
        # st.write(plt.gcf())


def test_das_gained():
    from .resources import df_passes, df_tracking
    df_passes = df_passes.copy()
    df_tracking = df_tracking.copy()

    ret_das_gained = accessible_space.get_das_gained(df_passes, df_tracking, respect_offside=False, use_event_coordinates_as_ball_position=False, use_event_team_as_team_in_possession=False, tracking_period_col=None)
    df_passes["DAS_gained"] = ret_das_gained.das_gained
    df_passes["AS_gained"] = ret_das_gained.as_gained
    df_passes["AS"] = ret_das_gained.acc_space
    df_passes["DAS"] = ret_das_gained.das
    df_passes["AS_reception"] = ret_das_gained.acc_space_reception
    df_passes["DAS_reception"] = ret_das_gained.das_reception
    df_passes["frame_index"] = ret_das_gained.frame_index
    df_passes["target_frame_index"] = ret_das_gained.target_frame_index

    assert df_passes.apply(lambda row: row["frame_index"] != row["target_frame_index"], axis=1).all()

    i_unsuccessful = df_passes["pass_outcome"] == 0
    i_successful = df_passes["pass_outcome"] == 1

    assert (df_passes.loc[i_unsuccessful, "DAS_gained"] < 0).all()
    assert (df_passes.loc[i_unsuccessful, "AS_gained"] < 0).all()

    assert (df_passes.loc[i_successful, "DAS_gained"] == df_passes.loc[i_successful, "DAS_reception"] - df_passes.loc[i_successful, "DAS"]).all()

    ret_das = accessible_space.get_dangerous_accessible_space(df_tracking, period_col=None, respect_offside=False)
    df_tracking["DAS"] = ret_das.das
    df_tracking["AS"] = ret_das.acc_space

    for _, p4ss in df_passes.iterrows():
        df_tracking_frame = df_tracking[df_tracking["frame_id"] == p4ss["frame_id"]]

        assert p4ss["DAS"] == df_tracking_frame["DAS"].iloc[0]
        assert p4ss["AS"] == df_tracking_frame["AS"].iloc[0]

        df_tracking_target_frame = df_tracking[df_tracking["frame_id"] == p4ss["target_frame_id"]]
        if p4ss["pass_outcome"] == 1:
            assert p4ss["DAS_reception"] == df_tracking_target_frame["DAS"].iloc[0]
            assert p4ss["AS_reception"] == df_tracking_target_frame["AS"].iloc[0]
        else:
            assert p4ss["DAS_reception"] == 0
            assert p4ss["AS_reception"] == 0


def test_das_gained_duplicate_frames():
    from .resources import df_passes, df_tracking
    df_passes = df_passes.copy()
    df_tracking = df_tracking.copy()
    df_passes2 = df_passes.copy()
    # df_passes2["y"] = df_passes2["y"]
    # df_passes2["x"] = df_passes2["x"]
    # df_passes2["y_target"] = df_passes2["y_target"]
    # df_passes2["x_target"] = df_passes2["x_target"]
    df_passes = pd.concat([df_passes, df_passes2], ignore_index=True)

    ret_das_gained = accessible_space.get_das_gained(df_passes, df_tracking, respect_offside=False, use_event_coordinates_as_ball_position=True, use_event_team_as_team_in_possession=True, tracking_period_col=None)
    df_passes["DAS_gained"] = ret_das_gained.das_gained
    df_passes["AS_gained"] = ret_das_gained.as_gained
    df_passes["AS"] = ret_das_gained.acc_space
    df_passes["DAS"] = ret_das_gained.das
    df_passes["AS_reception"] = ret_das_gained.acc_space_reception
    df_passes["DAS_reception"] = ret_das_gained.das_reception
    df_passes["frame_index"] = ret_das_gained.frame_index
    df_passes["target_frame_index"] = ret_das_gained.target_frame_index

    # assert top half = bottom half
    assert np.allclose(df_passes.iloc[:len(df_passes) // 2]["DAS_gained"], df_passes.iloc[len(df_passes) // 2:]["DAS_gained"], atol=1e-3)

    assert df_passes.apply(lambda row: row["frame_index"] != row["target_frame_index"], axis=1).all()

    i_unsuccessful = df_passes["pass_outcome"] == 0
    i_successful = df_passes["pass_outcome"] == 1

    assert (df_passes.loc[i_unsuccessful, "DAS_gained"] < 0).all()
    assert (df_passes.loc[i_unsuccessful, "AS_gained"] < 0).all()

    assert (df_passes.loc[i_successful, "DAS_gained"] == df_passes.loc[i_successful, "DAS_reception"] - df_passes.loc[i_successful, "DAS"]).all()


def test_sort_independence():
    from .resources import df_tracking, df_passes
    df_tracking = df_tracking.copy()
    df_passes = df_passes.copy()

    df_tracking = pd.concat([df_tracking, df_tracking], ignore_index=True)
    df_passes = pd.concat([df_passes, df_passes], ignore_index=True)

    # DAS
    df_tracking = df_tracking.sort_values("frame_id", ascending=True).reset_index(drop=True)
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, period_col=None, respect_offside=False)
    df_tracking["DAS"] = ret.das

    df_tracking_sorted = df_tracking.sort_values("frame_id", ascending=False).reset_index(drop=True)
    ret_sorted = accessible_space.get_dangerous_accessible_space(df_tracking_sorted, period_col=None, respect_offside=False)
    df_tracking_sorted["DAS"] = ret_sorted.das

    df_tracking_sorted = df_tracking_sorted.sort_values("frame_id", ascending=True).reset_index(drop=True)
    assert (df_tracking["DAS"] == df_tracking_sorted["DAS"]).all()

    # xC
    df_passes = df_passes.sort_values("frame_id", ascending=True).reset_index(drop=True)
    ret = accessible_space.get_expected_pass_completion(df_passes, df_tracking, respect_offside=False)
    df_passes["xc"] = ret.xc

    df_passes_sorted = df_passes.sort_values("frame_id", ascending=False).reset_index(drop=True)
    ret_sorted = accessible_space.get_expected_pass_completion(df_passes_sorted, df_tracking, respect_offside=False)
    df_passes_sorted["xc"] = ret_sorted.xc

    df_passes_sorted = df_passes_sorted.sort_values("frame_id", ascending=True).reset_index(drop=True)
    assert (df_passes["xc"] == df_passes_sorted["xc"]).all()

    # DAS Gained
    df_passes = df_passes.sort_values("frame_id", ascending=True).reset_index(drop=True)
    ret = accessible_space.get_das_gained(df_passes, df_tracking, tracking_period_col="period_id", respect_offside=False)
    df_passes["DAS Gained"] = ret.das_gained

    df_passes_sorted = df_passes.sort_values("frame_id", ascending=False).reset_index(drop=True)
    ret_sorted = accessible_space.get_das_gained(df_passes_sorted, df_tracking, tracking_period_col="period_id", respect_offside=False)
    df_passes_sorted["DAS Gained"] = ret_sorted.das_gained

    df_passes_sorted = df_passes_sorted.sort_values("frame_id", ascending=True).reset_index(drop=True)
    assert (df_passes["DAS Gained"] == df_passes_sorted["DAS Gained"]).all()


def test_chunk_wise_simulation():
    from .resources import df_tracking, df_passes
    df_tracking = df_tracking.copy()
    df_passes = df_passes.copy()

    F_tracking = len(df_tracking["frame_id"].unique())
    assert F_tracking > 1
    F_event = len(df_passes["frame_id"].unique())
    assert F_event > 1

    for chunk_size in [500, 1, 2, 3, None, 0, -1]:
        ret = accessible_space.get_expected_pass_completion(df_passes, df_tracking, respect_offside=False, chunk_size=chunk_size, additional_fields_to_return=["defense_cum_poss"])
        assert ret.xc.shape[0] == len(df_passes)
        assert ret.event_frame_index.shape[0] == len(df_passes)
        assert ret.tracking_frame_index.shape[0] == len(df_tracking)
        assert ret.simulation_result.defense_cum_poss.shape[0] == F_event
        ret_field = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, chunk_size=chunk_size, additional_fields_to_return=["defense_poss_density"])
        assert ret_field.acc_space.shape[0] == len(df_tracking)
        assert ret_field.das.shape[0] == len(df_tracking)
        assert ret_field.simulation_result.defense_poss_density.shape[0] == F_tracking


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_cum_prob_sum_is_1(_get_data):
    _, _, df_tracking = _get_data()
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, additional_fields_to_return=["attack_cum_prob", "defense_cum_prob", "cum_p0"])
    p_sum = ret.simulation_result.attack_cum_prob[0] + ret.simulation_result.defense_cum_prob[0] + ret.simulation_result.cum_p0[0]
    assert np.allclose(p_sum, 1)


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_simulation_result_dimensions(_get_data):
    _, _, df_tracking = _get_data()
    F = len(df_tracking["frame_id"].unique())
    P = len(df_tracking.loc[df_tracking["player_id"] != "ball", "player_id"].unique())
    n_angles = 24
    importlib.reload(accessible_space)
    importlib.reload(accessible_space)
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, n_angles=n_angles, additional_fields_to_return=accessible_space.ALL_OPTIONAL_FIELDS_TO_RETURN)

    fields = [
        ret.simulation_result.attack_poss_density,
        ret.simulation_result.defense_poss_density,
        ret.simulation_result.attack_cum_poss,
        ret.simulation_result.defense_cum_poss,
        ret.simulation_result.attack_prob_density,
        ret.simulation_result.defense_prob_density,
        ret.simulation_result.attack_cum_prob,
        ret.simulation_result.defense_cum_prob,
        ret.simulation_result.cum_p0,
        ret.simulation_result.p0_density,
        ret.simulation_result.x_grid,
        ret.simulation_result.y_grid,
    ]
    T = ret.simulation_result.attack_poss_density.shape[2]
    for field in fields:
        assert len(field.shape) == 3
        assert field.shape[0] == F
        assert field.shape[1] == n_angles
        assert field.shape[2] == T

    individual_fields = [
        ret.simulation_result.player_cum_prob,
        ret.simulation_result.player_cum_poss,
        ret.simulation_result.player_prob_density,
        ret.simulation_result.player_poss_density,
    ]
    for field in individual_fields:
        assert len(field.shape) == 4
        assert field.shape[0] == F
        assert field.shape[1] == P
        assert field.shape[2] == n_angles
        assert field.shape[3] == T

    assert len(ret.simulation_result.r_grid.shape) == 1
    assert ret.simulation_result.r_grid.shape[0] == T
    assert len(ret.simulation_result.phi_grid.shape) == 2
    assert ret.simulation_result.phi_grid.shape[0] == F
    assert ret.simulation_result.phi_grid.shape[1] == n_angles


### TODO Re-introduce after figuring out correct normalization
# def test_integrated_prob_density_sum_is_1():
#     _, _, df_tracking = _get_butterfly_data()
#     normalize2ret = {}
#     for normalize in [True]:
#         st.write(f"#### normalize={normalize}")
#         ret = accessible_space.get_dangerous_accessible_space(df_tracking, period_col=None)
#         r_grid = ret.simulation_result.r_grid
#         p_cum_att_from_density = scipy.integrate.cumulative_trapezoid(y=ret.simulation_result.attack_prob_density, x=r_grid[np.newaxis, np.newaxis, :], initial=0, axis=-1)
#         p_cum_def_from_density = scipy.integrate.cumulative_trapezoid(y=ret.simulation_result.defense_prob_density, x=r_grid[np.newaxis, np.newaxis, :], initial=0, axis=-1)
#         p0_cum_from_density = 1 + scipy.integrate.cumulative_trapezoid(y=ret.simulation_result.p0_density, x=r_grid[np.newaxis, np.newaxis, :], initial=0, axis=-1)
#
#         st.write("p_cum_att_from_density", p_cum_att_from_density.shape)
#         st.write(p_cum_att_from_density[0])
#         st.write("p_cum_def_from_density", p_cum_def_from_density.shape)
#         st.write(p_cum_def_from_density[0])
#         st.write("p0_cum_from_density", p0_cum_from_density.shape)
#         st.write(p0_cum_from_density[0])
#         st.write("ret.simulation_result.cum_p0", ret.simulation_result.cum_p0.shape)
#         st.write(ret.simulation_result.cum_p0[0])
#         st.write((p_cum_att_from_density + p_cum_def_from_density + p0_cum_from_density)[0])
#
#         for field in [
#             p_cum_att_from_density,
#             p_cum_def_from_density,
#             p0_cum_from_density,
#         ]:
#             assert np.all(field >= 0)
#             assert np.all(field <= 1)
#
#         normalize2ret[normalize] = ret
#
#     norm_res = normalize2ret[True].simulation_result
#     assert ((norm_res.p_cum_att_from_density + norm_res.p_cum_def_from_density + norm_res.p0_cum_from_density) == 1).all()


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_probability_ranges(_get_data):
    _, _, df_tracking = _get_data()
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, additional_fields_to_return=accessible_space.ALL_OPTIONAL_FIELDS_TO_RETURN)

    for p_cum in [
        ret.simulation_result.attack_cum_prob,
        ret.simulation_result.defense_cum_prob,
        ret.simulation_result.attack_cum_poss,
        ret.simulation_result.defense_cum_poss,
        ret.simulation_result.cum_p0,
        ret.simulation_result.player_cum_prob,
        ret.simulation_result.player_cum_poss,
    ]:
        assert (p_cum >= 0).all()
        assert (p_cum <= 1).all()

    dx = ret.simulation_result.r_grid[1] - ret.simulation_result.r_grid[0]
    for p_density in [
        ret.simulation_result.attack_prob_density,
        ret.simulation_result.defense_prob_density,
        ret.simulation_result.attack_poss_density,
        ret.simulation_result.defense_poss_density,
        ret.simulation_result.player_prob_density,
        ret.simulation_result.player_poss_density,
    ]:
        assert ((p_density * dx) >= 0).all()
        assert ((p_density * dx) <= 1).all()

    assert ((ret.simulation_result.p0_density * dx) <= 0).all()
    assert ((ret.simulation_result.p0_density * dx) >= -1).all()


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_das_is_smaller_than_as(_get_data):
    _, _, df_tracking = _get_data()
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None)
    assert (ret.das <= ret.acc_space).all()


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_player_level_consistent_with_team_level(_get_data):
    _, _, df_tracking = _get_data()
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, additional_fields_to_return=accessible_space.ALL_OPTIONAL_FIELDS_TO_RETURN)

    df_tracking["frame_index"], df_tracking["player_index"] = ret.frame_index, ret.player_index

    i_ball = df_tracking["team_id"] == "ball"
    i_att = df_tracking["team_id"] == df_tracking["team_in_possession"]
    attacking_player_indices = df_tracking.loc[~i_ball & i_att, "player_index"].dropna().astype(int).unique()
    defending_player_indices = df_tracking.loc[~i_ball & ~i_att, "player_index"].dropna().astype(int).unique()

    p_density_att_from_players = ret.simulation_result.player_prob_density[0, attacking_player_indices, :, :].sum(axis=0)
    p_density_def_from_players = ret.simulation_result.player_prob_density[0, defending_player_indices, :, :].sum(axis=0)
    assert np.allclose(p_density_att_from_players, ret.simulation_result.attack_prob_density[0])
    assert np.allclose(p_density_def_from_players, ret.simulation_result.defense_prob_density[0])

    poss_density_att_from_players = ret.simulation_result.player_poss_density[0, attacking_player_indices, :, :].max(axis=0)
    poss_density_def_from_players = ret.simulation_result.player_poss_density[0, defending_player_indices, :, :].max(axis=0)
    assert np.allclose(poss_density_att_from_players, ret.simulation_result.attack_poss_density[0])
    assert np.allclose(poss_density_def_from_players, ret.simulation_result.defense_poss_density[0])

    p_cum_att_from_players = ret.simulation_result.player_cum_prob[0, attacking_player_indices, :, :].sum(axis=0)
    p_cum_def_from_players = ret.simulation_result.player_cum_prob[0, defending_player_indices, :, :].sum(axis=0)
    assert np.allclose(p_cum_att_from_players, ret.simulation_result.attack_cum_prob[0])
    assert np.allclose(p_cum_def_from_players, ret.simulation_result.defense_cum_prob[0])

    poss_cum_att_from_players = ret.simulation_result.player_cum_poss[0, attacking_player_indices, :, :].max(axis=0)
    poss_cum_def_from_players = ret.simulation_result.player_cum_poss[0, defending_player_indices, :, :].max(axis=0)
    assert np.allclose(poss_cum_att_from_players, ret.simulation_result.attack_cum_poss[0])
    assert np.allclose(poss_cum_def_from_players, ret.simulation_result.defense_cum_poss[0])


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_infer_playing_direction(_get_data):
    _, _, df_tracking = _get_data()
    df_tracking["playing_direction"] = accessible_space.infer_playing_direction(df_tracking, period_col=None, ball_team="BALL")
    assert (df_tracking["playing_direction"] == 1).all()

@pytest.mark.parametrize("df_tracking,period_col,exception,exception_message_substring", [
    (pd.DataFrame(), None, KeyError, "Missing columns"),
    (pd.DataFrame({"frame_id": [1, 2]}), None, KeyError, "Missing columns"),
    (pd.DataFrame({"frame_id": [1, 2], "team_id": ["H", "A"], "x": [0, 0]}), None, KeyError, "Missing column"),
    (pd.DataFrame({"frame_id": [1, 2], "team_id": ["H", "A"], "x": [0, 0], "team_in_possession": ["H", "H"]}), "period_col", KeyError, "Missing column"),
])
def test_bad_data_infer_playing_direction(df_tracking, period_col, exception, exception_message_substring):
    with pytest.raises(exception, match=exception_message_substring):
        accessible_space.infer_playing_direction(df_tracking, period_col=period_col)


@pytest.mark.parametrize("df_tracking,exception,exception_message_substring", [
    (pd.DataFrame({"frame_id": [1, 2], "team_id": ["H", "A"], "x": [0, 0], "team_in_possession": ["H", "H"]}), KeyError, "specify period_col or set to None if your data has no separate periods"),
])
def test_bad_data_infer_playing_direction_with_default_period(df_tracking, exception, exception_message_substring):
    with pytest.raises(exception, match=exception_message_substring):
        accessible_space.infer_playing_direction(df_tracking)


### TODO: Test is not working due to normalization problems
# @pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
# def test_poss_never_below_prob(_get_data):
#     _, _, df_tracking = _get_data()
#     ret = accessible_space.get_dangerous_accessible_space(df_tracking, period_col=None, normalize=True)
#
#     for (prob, poss) in [
#         # (ret.simulation_result.player_prob_density, ret.simulation_result.player_poss_density),
#         # (ret.simulation_result.player_cum_prob, ret.simulation_result.player_cum_poss),
#         # (ret.simulation_result.attack_prob_density, ret.simulation_result.attack_poss_density),
#         # (ret.simulation_result.defense_prob_density, ret.simulation_result.defense_poss_density),
#         (ret.simulation_result.attack_cum_prob, ret.simulation_result.attack_cum_poss),
#         (ret.simulation_result.defense_cum_prob, ret.simulation_result.defense_cum_poss),
#     ]:
#         assert np.all(prob <= poss)  # all smaller or equal


@pytest.mark.parametrize("x_min,x_max,y_min,y_max,_get_data", [
    (-52.5, 52.5, -34, 34, _get_butterfly_data),
    (0, 105, 0, 68, _get_butterfly_data),
    (-1, 1, -1, 1, _get_butterfly_data),
    (1500, 2000, 1500, 1600, _get_butterfly_data),
    (20, 200, 20, 200, _get_butterfly_data),
    (-52.5, 52.5, -34, 34, _get_butterfly_data_with_nans),
    (0, 105, 0, 68, _get_butterfly_data_with_nans),
    (-1, 1, -1, 1, _get_butterfly_data_with_nans),
    (1500, 2000, 1500, 1600, _get_butterfly_data_with_nans),
    (20, 200, 20, 200, _get_butterfly_data_with_nans),
])
def test_pitch_clipping(x_min, x_max, y_min, y_max, _get_data):
    _, _, df_tracking = _get_data()
    density_fields = [field for field in accessible_space.ALL_OPTIONAL_FIELDS_TO_RETURN if "density" in field]
    importlib.reload(accessible_space)
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, additional_fields_to_return=accessible_space.ALL_OPTIONAL_FIELDS_TO_RETURN)
    cropped_result = accessible_space.clip_simulation_result_to_pitch(ret.simulation_result, x_min, x_max, y_min, y_max)

    x_grid = ret.simulation_result.x_grid
    y_grid = ret.simulation_result.y_grid
    i_in_pitch = (x_grid >= x_min) & (x_grid <= x_max) & (y_grid >= y_min) & (y_grid <= y_max)

    for field_str in density_fields:
        field = getattr(ret.simulation_result, field_str)
        cropped_field = getattr(cropped_result, field_str)
        i_in_pitch_player = np.repeat(i_in_pitch[:, np.newaxis, :, :], field.shape[1], axis=1) if len(field.shape) == 4 else i_in_pitch

        assert np.all(field[i_in_pitch_player] == cropped_field[i_in_pitch_player])
        assert np.all(cropped_field[~i_in_pitch_player] == 0)

    for field_str in [
        "attack_cum_prob", "defense_cum_prob", "attack_cum_poss", "defense_cum_poss", "cum_p0", "player_cum_prob",
        "player_cum_poss"
    ]:
        cum_field = getattr(ret.simulation_result, field_str)
        cropped_field = getattr(cropped_result, field_str)
        i_in_pitch_field = i_in_pitch if len(cum_field.shape) == 3 else np.repeat(i_in_pitch[:, np.newaxis, :, :], cum_field.shape[1], axis=1)

        assert np.all(cum_field[i_in_pitch_field] == cropped_field[i_in_pitch_field])

        def array_to_list_of_non_nan_1d_slices(array):
            reshaped_array = array.reshape(-1, array.shape[-1])
            list_of_arrays = [row[~np.isnan(row)] for row in reshaped_array]
            return list_of_arrays

        field_outside_pitch = np.where(~i_in_pitch_field, cropped_field, np.nan)
        for time_slice_outside_pitch in array_to_list_of_non_nan_1d_slices(field_outside_pitch):
            assert len(np.unique(time_slice_outside_pitch)) == 1


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_surface_integration(_get_data):
    _, _, df_tracking = _get_data()
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, additional_fields_to_return=[field for field in accessible_space.ALL_OPTIONAL_FIELDS_TO_RETURN if "density" in field])
    areas = accessible_space.integrate_surfaces(ret.simulation_result)
    for field in areas:
        assert np.all(field >= 0)
        assert np.all(field <= 105*68)


@pytest.mark.parametrize("df_tracking,exception,exception_message_substring", [
    (pd.DataFrame(), KeyError, "Missing columns in tracking data"),
    (pd.DataFrame({"frame_id": [1, 2]}), KeyError, "Missing columns in tracking data"),
    (pd.DataFrame({"frame_id": [1, 2], "player_id": ["a", "b"], "team_id": ["H", "A"], "x": [0, 0], "vx": [0, 0]}), KeyError, "Missing columns in tracking data: y_col='y', vy_col='vy'"),
    (pd.DataFrame({"frame_id": [1, 2], "player_id": ["a", "b"], "team_id": ["H", "A"], "x": [0, 0], "y": [0, 0], "vx": [0, 0]}), KeyError, "Missing columns in tracking data: vy_col='vy'"),
    (pd.DataFrame({"frame_id": [1, 2], "player_id": ["a", "b"], "team_id": ["H", "A"], "x": [0, 0], "y": [0, 0], "vx": [0, 0], "vy": [0, 0]}), KeyError, "Missing column in tracking data: team_in_possession_col='team_in_possession'"),
    (pd.DataFrame({"frame_id": [1, 2], "player_id": ["a", "b"], "team_id": ["H", "A"], "x": [0, 0], "y": [0, 0], "vx": [0, 0], "vy": [0, 0], "team_in_possession": ["H", "H"]}), ValueError, "Ball flag ball_tracking_player_id='ball' does not exist in column "),
])
def test_bad_data_das(df_tracking, exception, exception_message_substring):
    with pytest.raises(exception, match=exception_message_substring):
        accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None)

#     df_tracking =
@pytest.mark.parametrize("df_passes,df_tracking,exception,use_event_ball_pos,exception_message_substring", [
    (pd.DataFrame({"frame_id": [1, 1], "player_id": ["a", "a"], "team_id": ["H", "H"], "x": [0, 0], "x_target": [1, 1], "y": [2, 2], "y_target": [3, 3]}), pd.DataFrame(), KeyError, True, "Missing columns in df_tracking"),
    (pd.DataFrame(), pd.DataFrame({"frame_id": [1, 1, 1, 1], "player_id": ["a", "ball", "b", "c"], "team_id": ["H", None, "A", "H"], "x": [0, 0, 1, 2], "y": [0, 0, 1, 2], "vx": [0, 0, 1, 2], "vy": [0, 0, 1, 2], "team_in_possession": ["H", "H", "H", "H"]}), KeyError, True, "Missing columns in df_passes"),
    (pd.DataFrame({"frame_id": [1, 1], "player_id": ["a", "a"], "team_id": ["H", "H"], "x": [0, 0], "x_target": [1, 1], "y": [2, 2], "y_target": [3, 3]}), pd.DataFrame({"frame_id": [1, 1], "player_id": ["a", "b"], "team_id": ["H", "A"], "x": [0, 0], "y": [0, 0], "vx": [0, 0], "vy": [0, 0], "team_in_possession": ["H", "H"]}), ValueError, True, "Ball flag ball_tracking_player_id='ball' does not exist in column"),
])
def test_bad_data_xc(df_passes, df_tracking, use_event_ball_pos, exception, exception_message_substring):
    with pytest.raises(exception, match=exception_message_substring):
        accessible_space.get_expected_pass_completion(df_passes, df_tracking, respect_offside=False)


def test_duplicate_frames():
    df_passes = pd.DataFrame({"frame_id": [5, 5], "player_id": ["a", "a"], "team_id": ["H", "H"], "x": [0, 0], "x_target": [1, 1], "y": [2, 2], "y_target": [3, 3]})
    df_tracking = pd.DataFrame({"frame_id": [5, 5, 5, 5], "player_id": ["a", "ball", "b", "c"], "team_id": ["H", None, "A", "H"], "x": [0, 0, 1, 2], "y": [0, 0, 1, 2], "vx": [0, 0, 1, 2], "vy": [0, 0, 1, 2], "team_in_possession": ["H", "H", "H", "H"]})
    ret = accessible_space.get_expected_pass_completion(df_passes, df_tracking, respect_offside=False, additional_fields_to_return=["attack_poss_density"])
    assert len(ret.xc) == len(df_passes)
    assert len(ret.event_frame_index) == len(df_passes)
    assert len(ret.tracking_frame_index) == len(df_tracking)
    assert len(ret.tracking_player_index) == len(df_tracking)

    df_passes["xC"] = ret.xc
    df_passes["frame_index"] = ret.event_frame_index
    df_tracking["frame_index"] = ret.tracking_frame_index
    df_tracking["player_index"] = ret.tracking_player_index

    assert len(df_passes["frame_index"].unique()) == len(df_passes)
    assert ret.simulation_result.attack_poss_density.shape[0] == len(df_passes)


def test_minimal_das_runs_error_free():
    df_tracking = pd.DataFrame({"frame_id": [1, 1, 1, 1], "player_id": ["a", "ball", "b", "c"], "team_id": ["H", None, "A", "H"], "x": [0, 0, 1, 2], "y": [0, 0, 1, 2], "vx": [0, 0, 1, 2], "vy": [0, 0, 1, 2], "team_in_possession": ["H", "H", "H", "H"]})
    accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None)


def test_minimal_das_player_runs_error_free():
    df_tracking = pd.DataFrame({"frame_id": [1, 1, 1, 1], "player_id": ["a", "ball", "b", "c"], "team_id": ["H", None, "A", "H"], "x": [0, 0, 1, 2], "y": [0, 0, 1, 2], "vx": [0, 0, 1, 2], "vy": [0, 0, 1, 2], "team_in_possession": ["H", "H", "H", "H"]})
    accessible_space.get_individual_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None)


def test_minimal_xc_runs_error_free():
    df_tracking = pd.DataFrame({"frame_id": [1, 1, 1, 1], "player_id": ["a", "ball", "b", "c"], "team_id": ["H", None, "A", "H"], "x": [0, 0, 1, 2], "y": [0, 0, 1, 2], "vx": [0, 0, 1, 2], "vy": [0, 0, 1, 2], "team_in_possession": ["H", "H", "H", "H"]})
    df_passes = pd.DataFrame({"frame_id": [1, 1], "player_id": ["a", "a"], "team_id": ["H", "H"], "x": [0, 0], "x_target": [1, 1], "y": [2, 2], "y_target": [3, 3]})
    accessible_space.get_expected_pass_completion(df_passes, df_tracking, respect_offside=False)



@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_fields_to_return_are_present(_get_data):
    _, _, df_tracking = _get_data()

    for field in [
        "attack_cum_prob",
        "attack_cum_poss",
        "attack_prob_density",
        "attack_poss_density",
        "defense_cum_prob",
        "defense_cum_poss",
        "defense_prob_density",
        "defense_poss_density",
        "cum_p0",
        "p0_density",
        "player_cum_prob",
        "player_cum_poss",
        "player_prob_density",
        "player_poss_density",
    ]:
        ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, additional_fields_to_return=[field])
        assert getattr(ret.simulation_result, field) is not None


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_fields_to_return_others_are_not_present(_get_data):
    _, _, df_tracking = _get_data()

    all_fields = [
        "attack_cum_prob",
        "attack_cum_poss",
        "attack_prob_density",
        "attack_poss_density",
        "defense_cum_prob",
        "defense_cum_poss",
        "defense_prob_density",
        "defense_poss_density",
        "cum_p0",
        "p0_density",
        "player_cum_prob",
        "player_cum_poss",
        "player_prob_density",
        "player_poss_density",
    ]
    das_fields = ["attack_poss_density"]

    for field in all_fields:
        ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, additional_fields_to_return=[field])

        none_fields = [ret_field for ret_field in all_fields if getattr(ret.simulation_result, ret_field) is None]
        present_fields = [ret_field for ret_field in all_fields if isinstance(getattr(ret.simulation_result, ret_field), np.ndarray)]
        remaining_fields = [ret_field for ret_field in all_fields if ret_field not in none_fields and ret_field not in present_fields]

        expected_fields = {field}.union(set(das_fields))

        assert len(remaining_fields) == 0
        assert set(present_fields) == expected_fields

    for field1 in all_fields:
        for field2 in all_fields:
            if field1 == field2:
                continue
            ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, additional_fields_to_return=[field1, field2])

            none_fields = [ret_field for ret_field in all_fields if getattr(ret.simulation_result, ret_field) is None]
            present_fields = [ret_field for ret_field in all_fields if isinstance(getattr(ret.simulation_result, ret_field), np.ndarray)]
            remaining_fields = [ret_field for ret_field in all_fields if ret_field not in none_fields and ret_field not in present_fields]

            expected_fields = {field1, field2}.union(set(das_fields))

            assert len(remaining_fields) == 0
            assert set(present_fields) == expected_fields
            for present_field in present_fields:
                field_data = getattr(ret.simulation_result, present_field)
                assert np.all(~np.isnan(field_data))
                assert np.any(field_data != 0)


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_surface_plot(_get_data):
    _, _, df_tracking = _get_data()
    ret = accessible_space.get_dangerous_accessible_space(df_tracking, respect_offside=False, period_col=None, additional_fields_to_return=[field for field in accessible_space.ALL_OPTIONAL_FIELDS_TO_RETURN if "density" in field])
    matplotlib.use("Agg")

    def _plot():
        plt.figure()
        plt.xlim([-52.5, 52.5])
        plt.ylim([-34, 34])
        plt.scatter(df_tracking["x"], df_tracking["y"], color=df_tracking["player_color"])

    import streamlit as st
    _plot()
    accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "attack_poss_density")
    st.write("attack_poss_density")
    st.write(plt.gcf())
    plt.close()
    _plot()
    accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "attack_prob_density")
    st.write("attack_prob_density")
    st.write(plt.gcf())
    plt.close()
    _plot()
    accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "defense_poss_density")
    st.write("defense_poss_density")
    st.write(plt.gcf())
    plt.close()
    _plot()
    accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "defense_prob_density")
    st.write("defense_prob_density")
    st.write(plt.gcf())
    plt.close()
    _plot()
    accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "player_prob_density", player_index=0)
    st.write("player_prob_density")
    st.write(plt.gcf())
    plt.close()
    _plot()
    accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "player_poss_density", player_index=0)
    st.write("player_poss_density")
    st.write(plt.gcf())
    plt.close()

    with pytest.raises(ValueError, match=f"no player index is given"):
        _plot()
        accessible_space.plot_expected_completion_surface(ret.simulation_result, 0, "player_prob_density")

    plt.close()


@pytest.mark.parametrize("_get_data", [_get_butterfly_data, _get_butterfly_data_with_nans])
def test_additional_defender_decreases_as_and_additional_attacker_increases_as(_get_data):
    _, _, df_tracking = _get_data()

    def get_as_and_das(_df_tracking):
        ret_baseline = accessible_space.get_dangerous_accessible_space(
            _df_tracking, respect_offside=False, infer_attacking_direction=False, attacking_direction_col="attacking_direction",
        )
        return ret_baseline.acc_space.iloc[0], ret_baseline.das.iloc[0], ret_baseline

    baseline_as, baseline_das, baseline_result = get_as_and_das(df_tracking)

    # plot
    # plt.figure()
    # plt.xlim([-52.5, 52.5])
    # plt.ylim([-34, 34])
    # plt.scatter(df_tracking["x"], df_tracking["y"], color=df_tracking["player_color"])
    # plt.scatter(df_tracking["x"].iloc[0], df_tracking["y"].iloc[0], color="red", label="ball")
    # accessible_space.plot_expected_completion_surface(baseline_result.simulation_result, 0, "attack_poss_density")
    # st.write(plt.gcf())

    defending_team = [team for team in df_tracking["team_id"].unique() if team != df_tracking["team_in_possession"].iloc[0]][0]
    attacking_team = df_tracking["team_in_possession"].iloc[0]

    for new_x in [0, 10, 53]:
        for new_y in [-10, 40]:
            for new_vx in [0, -20]:
                for new_vy in [0, -2]:
                    df_tracking_extra_defender = df_tracking.copy()
                    extra_defender_data = {
                        "frame_id": df_tracking["frame_id"].iloc[0],
                        "player_id": "extra_player",
                        "x": new_x, "y": new_y, "vx": new_vx, "vy": new_vy,
                        "team_id": defending_team,
                        "player_color": None,
                        "attacking_direction": df_tracking["attacking_direction"].iloc[0],
                        "team_in_possession": df_tracking["team_in_possession"].iloc[0],
                        "player_in_possession": df_tracking["player_in_possession"].iloc[0],
                    }
                    df_tracking_extra_defender.loc[len(df_tracking_extra_defender)] = pd.Series(extra_defender_data)

                    as_with_extra_defender, das_with_extra_defender, _ = get_as_and_das(df_tracking_extra_defender)
                    assert as_with_extra_defender <= baseline_as
                    assert das_with_extra_defender <= baseline_das

                    df_tracking_extra_attacker = df_tracking.copy()
                    extra_attacker_data = {
                        "frame_id": df_tracking["frame_id"].iloc[0],
                        "player_id": "extra_player",
                        "x": new_x, "y": new_y, "vx": new_vx, "vy": new_vy,
                        "team_id": attacking_team,
                        "player_color": None,
                        "attacking_direction": df_tracking["attacking_direction"].iloc[0],
                        "team_in_possession": df_tracking["team_in_possession"].iloc[0],
                        "player_in_possession": df_tracking["player_in_possession"].iloc[0],
                    }
                    # df_tracking_extra_attacker = df_tracking_extra_attacker.append(pd.Series(extra_attacker_data), ignore_index=True)
                    df_tracking_extra_attacker.loc[len(df_tracking_extra_attacker)] = pd.Series(extra_attacker_data)
                    as_with_extra_attacker, das_with_extra_attacker, _ = get_as_and_das(df_tracking_extra_attacker)
                    assert as_with_extra_attacker >= baseline_as, f"new_x={new_x}, new_y={new_y}, new_vx={new_vx} new_vy={new_vy}"
                    assert das_with_extra_attacker >= baseline_das



import pytest
import accessible_space
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import xmltodict
import requests


def _get_metrica_data(
    dataset_nr=1, new_team_col="TEAM", new_player_col="PLAYER", new_x_col="X", new_y_col="Y",
    passer_team_col="passer_team", receiver_team_col="receiver_team",
):
    # @st.cache_resource
    def get_events(dataset_nr):
        metrica_open_data_base_dir = "https://raw.githubusercontent.com/metrica-sports/sample-data/refs/heads/master/data"
        if dataset_nr in [1, 2]:
            df = pd.read_csv(f"{metrica_open_data_base_dir}/Sample_Game_{dataset_nr}/Sample_Game_{dataset_nr}_RawEventsData.csv")
            df["body_part_type"] = df["Subtype"].where(df["Subtype"].isin(["HEAD"]), None)
            df["set_piece_type"] = df["Subtype"].where(
                df["Subtype"].isin(["THROW IN", "GOAL KICK", "FREE KICK", "CORNER KICK"]), None).map(
                lambda x: x.replace(" ", "_") if x is not None else None
            )
            df["Type"] = df["Type"].str.replace(" ", "_")
            df["Start X"] = (df["Start X"] - 0.5) * 105
            df["Start Y"] = -(df["Start Y"] - 0.5) * 68
            df["End X"] = (df["End X"] - 0.5) * 105
            df["End Y"] = -(df["End Y"] - 0.5) * 68
            df = df.rename(columns={
                "Type": "event_type",
                "Period": "period_id",
                "Team": "team_id",
                "From": "player_id",
                "To": "receiver_player_id",
                "Start X": "coordinates_x",
                "Start Y": "coordinates_y",
                "End X": "end_coordinates_x",
                "End Y": "end_coordinates_y",
                "Start Frame": "frame_id",
                "End Frame": "end_frame_id",
            })
            player_id_to_column_id = {}
            column_id_to_team_id = {}
            for team_id in df["team_id"].unique():
                df_players = df[df["team_id"] == team_id]
                team_player_ids = set(df_players["player_id"].dropna().tolist() + df_players["receiver_player_id"].dropna().tolist())
                player_id_to_column_id.update({player_id: f"{team_id.lower().strip()}_{player_id.replace('Player', '').strip()}" for player_id in team_player_ids})
                column_id_to_team_id.update({player_id_to_column_id[player_id]: team_id for player_id in team_player_ids})


            df["PLAYER"] = df["player_id"].map(player_id_to_column_id)
            df["RECEIVER"] = df["receiver_player_id"].map(player_id_to_column_id)
            df["RECEIVING_TEAM"] = df["receiver_player_id"].map(column_id_to_team_id)

            df["tmp_next_player"] = df["player_id"].shift(-1)
            df["tmp_next_team"] = df["team_id"].shift(-1)
            df["tmp_receiver_player"] = df["receiver_player_id"].where(df["receiver_player_id"].notna(),
                                                                       df["tmp_next_player"])
            df["tmp_receiver_team"] = df["tmp_receiver_player"].map(column_id_to_team_id)

            df["is_successful"] = df["tmp_receiver_team"] == df["team_id"]

            df["is_pass"] = (df["event_type"].isin(["PASS", "BALL_LOST", "BALL_OUT"])) \
                            & (~df["Subtype"].isin(["CLEARANCE", "HEAD-CLEARANCE", "HEAD-INTERCEPTION-CLEARANCE"])) \
                            & (df["frame_id"] != df["end_frame_id"])

            df["is_high"] = df["Subtype"].isin([
                "CROSS",
                # "CLEARANCE",
                "CROSS-INTERCEPTION",
                # "HEAD-CLEARANCE",
                # "HEAD-INTERCEPTION-CLEARANCE"
            ])

            #     df_passes["xc"], _, _ = dangerous_accessible_space.get_expected_pass_completion(
            #         df_passes, df_tracking, event_frame_col="td_frame", tracking_frame_col="frame", event_start_x_col="start_x",
            #         event_start_y_col="start_y", event_end_x_col="end_x", event_end_y_col="end_y",
            #         event_player_col="tracking_player_id",
            #     )

            return df.drop(columns=["tmp_next_player", "tmp_next_team", "tmp_receiver_player", "tmp_receiver_team"])
        else:
            # json_data = json.loads(open(f"{metrica_open_data_base_dir}/Sample_Game_3/Sample_Game_3_events.json"))
            json_data = requests.get(f"{metrica_open_data_base_dir}/Sample_Game_3/Sample_Game_3_events.json").json()

            df = pd.json_normalize(json_data["data"])

            # expanded_df = pd.DataFrame(df['subtypes'].apply(pd.Series))
            expanded_df = pd.DataFrame(df['subtypes'].apply(
                lambda x: pd.Series(x, dtype='float64') if x is None or isinstance(x, float) and pd.isna(x) else pd.Series(x)
            ))
            expanded_df.columns = [f'subtypes.{col}' for col in expanded_df.columns]

            new_dfs = []
            for expanded_col in expanded_df.columns:
                expanded_df2 = pd.json_normalize(expanded_df[expanded_col])
                expanded_df2.columns = [f'{expanded_col}.{col}' for col in expanded_df2.columns]
                new_dfs.append(expanded_df2)

            expanded_df = pd.concat(new_dfs, axis=1)

            df = pd.concat([df, expanded_df], axis=1)

            i_subtypes_nan = ~df["subtypes.name"].isna()
            i_subtypes_0_nan = ~df["subtypes.0.name"].isna()

            # check if the True's are mutually exclusive
            assert not (i_subtypes_nan & i_subtypes_0_nan).any()

            df.loc[i_subtypes_nan, "subtypes.0.name"] = df.loc[i_subtypes_nan, "subtypes.name"]
            df.loc[i_subtypes_nan, "subtypes.0.id"] = df.loc[i_subtypes_nan, "subtypes.id"]
            df = df.drop(columns=["subtypes.name", "subtypes.id", "subtypes"])
            subtype_cols = [col for col in df.columns if col.startswith("subtypes.") and col.endswith("name")]

            player2team = df[['from.id', 'team.id']].set_index('from.id')['team.id'].to_dict()
            df["receiver_team_id"] = df["to.id"].map(player2team)
            # df["tmp_next_player"] = df["player_id"].shift(-1)
            # df["tmp_next_team"] = df["team_id"].shift(-1)
            # df["tmp_receiver_player"] = df["receiver_player_id"].where(df["receiver_player_id"].notna(), df["tmp_next_player"])
            # df["tmp_receiver_team"] = df["tmp_receiver_player"].map(column_id_to_team_id)
            df["success"] = df["receiver_team_id"] == df["team.id"]

            df["success"] = df["success"].astype(bool)

            df["is_pass"] = (df["type.name"].isin(["PASS", "BALL LOST", "BALL OUT"])) \
                            & ~df[subtype_cols].isin(["CLEARANCE"]).any(axis=1) \
                            & (df["start.frame"] != df["end.frame"])

            # df[df[['Name', 'Age']].isin(['Alice', 30]).any(axis=1)]
            df["is_high"] = df[subtype_cols].isin(["CROSS"]).any(axis=1)

            df = df.rename(columns={
                "type.name": "event_type",
                "from.id": "player_id",
                "team.id": "team_id",
                "to.id": "receiver_player_id",
                "period": "period_id",
                "start.frame": "frame_id",
                "end.frame": "end_frame_id",
                "start.x": "coordinates_x",
                "start.y": "coordinates_y",
                "end.x": "end_coordinates_x",
                "end.y": "end_coordinates_y",
            }).drop(columns=[
                "to",
            ])
            df["coordinates_x"] = (df["coordinates_x"] - 0.5) * 105
            df["coordinates_y"] = (df["coordinates_y"] - 0.5) * 68
            df["end_coordinates_x"] = (df["end_coordinates_x"] - 0.5) * 105
            df["end_coordinates_y"] = (df["end_coordinates_y"] - 0.5) * 68

            meta_data = xmltodict.parse(
                requests.get(f"{metrica_open_data_base_dir}/Sample_Game_3/Sample_Game_3_metadata.xml").text)

            df_player = pd.json_normalize(meta_data, record_path=["main", "Metadata", "Players", "Player"])
            player2team = df_player[["@id", "@teamId"]].set_index("@id")["@teamId"].to_dict()
            df["team_id"] = df["player_id"].map(player2team)

            return df

    # @st.cache_resource
    def _get():
        #                             https://raw.githubusercontent.com/metrica-sports/sample-data/refs/heads/master/data/Sample_Game_1/Sample_Game_1_RawEventsData.csv
        metrica_open_data_base_dir = "https://raw.githubusercontent.com/metrica-sports/sample-data/refs/heads/master/data"
        # home_data=f"https://raw.githubusercontent.com/metrica-sports/sample-data/master/data/Sample_Game_{dataset_nr}/Sample_Game_{dataset_nr}_RawTrackingData_Home_Team.csv"
        home_data_url = f"{metrica_open_data_base_dir}/Sample_Game_{dataset_nr}/Sample_Game_{dataset_nr}_RawTrackingData_Home_Team.csv"
        # away_data=f"https://raw.githubusercontent.com/metrica-sports/sample-data/master/data/Sample_Game_{dataset_nr}/Sample_Game_{dataset_nr}_RawTrackingData_Away_Team.csv"
        away_data_url = f"{metrica_open_data_base_dir}/Sample_Game_{dataset_nr}/Sample_Game_{dataset_nr}_RawTrackingData_Away_Team.csv"
        event_url = f"{metrica_open_data_base_dir}/Sample_Game_{dataset_nr}/Sample_Game_{dataset_nr}_RawEventsData.csv"
        # with st.spinner(f"Loading data 1/3 from {event_url}"):
        df_events = pd.read_csv(event_url)
        # with st.spinner(f"Loading data 2/3 from {home_data_url}"):
        df_home = pd.read_csv(home_data_url, skiprows=2)
        # with st.spinner(f"Loading data 3/3 from {away_data_url}"):
        df_away = pd.read_csv(away_data_url, skiprows=2)
        return df_home, df_away, df_events

    df_home, df_away, df_events = _get()
    df_events = get_events(dataset_nr)
    df_passes = df_events[df_events["is_pass"]].copy()

    # st.write("df_events")
    # st.write(df_events)

    df_tracking = []
    for team, df_team in [("Home", df_home), ("Away", df_away)]:
        x_cols = [col for col in df_team.columns if col.startswith("Player") or col.startswith("Ball")]
        y_cols = [col for col in df_team.columns if col.startswith("Unnamed")]
        player_cols = x_cols

        df_tracking.append(accessible_space.per_object_frameify_tracking_data(
            df_team,
            frame_col="Frame",
            coordinate_cols=[[x, y] for x, y in zip(x_cols, y_cols)],
            players=player_cols,
            player_to_team={player: team for player in player_cols},
            new_coordinate_cols=[new_x_col, new_y_col],
            new_team_col=new_team_col,
            new_player_col=new_player_col,
        ))
    df_tracking = pd.concat(df_tracking)
    df_tracking = df_tracking.drop_duplicates(subset=["Frame", new_player_col], keep="first")
    i_ball = df_tracking[new_player_col] == "Ball"
    df_tracking.loc[i_ball, new_team_col] = None

    df_tracking["vx"] = 0  # df["x"].diff() / 25  # v not present -> allow infer?
    df_tracking["vy"] = 0  # df["y"].diff() / 25  # TODO check v correctness in validation

    # i_pass = (df_events["Type"].isin(["PASS", "BALL LOST", "BALL OUT"])) \
    #          & (~df_events["Subtype"].isin(["CLEARANCE", "HEAD-CLEARANCE", "HEAD-INTERCEPTION-CLEARANCE"]))

    # df_passes = df_events[i_pass]
    player2team = df_tracking[[new_player_col, new_team_col]].set_index(new_player_col)[new_team_col].to_dict()
    df_passes[passer_team_col] = df_passes["player_id"].map(player2team)
    df_passes[receiver_team_col] = df_passes["receiver_player_id"].map(player2team)

    # for x_col in ["coordinates_x"]:
    #     df_passes[x_col] = (df_passes[x_col] - 0.5) * 105
    # for y_col in ["coordinates_y"]:
    #     df_passes[y_col] = (df_passes[y_col] - 0.5) * 68
    for x_col in [new_x_col]:
        df_tracking[x_col] = (df_tracking[x_col] - 0.5) * 105
    for y_col in [new_y_col]:
        df_tracking[y_col] = (df_tracking[y_col] - 0.5) * 68 * -1

    df_passes["is_successful"] = df_passes[passer_team_col] == df_passes[receiver_team_col]
    # st.write("df_passes")
    # st.write(df_passes)
    # st.write("Success rate", df_passes["is_successful"].mean())
    assert df_passes["is_successful"].mean() < 1.0

    return df_passes, df_tracking


@pytest.mark.parametrize("dataset_nr", [1, 2])
def test_real_world_data(dataset_nr):
    df_passes, df_tracking = _get_metrica_data(dataset_nr)

# KeyError: "Missing columns in df_passes: event_frame_col='Start Frame', event_start_x_col='Start X', event_start_y_col='Start Y', event_end_x_col='End X', event_end_y_col='End Y', event_team_col='Team', event_player_col='From'."

    ret = accessible_space.get_expected_pass_completion(
        df_passes=df_passes, df_tracking=df_tracking, tracking_frame_col="Frame", event_frame_col="frame_id",
        event_start_x_col="coordinates_x", event_end_x_col="end_coordinates_x", event_start_y_col="coordinates_y", event_end_y_col="end_coordinates_y",
        event_player_col="player_id", use_event_coordinates_as_ball_position=True, ball_tracking_player_id="Ball",
        event_team_col="passer_team", tracking_x_col="X", tracking_y_col="Y", tracking_team_col="TEAM",
        tracking_player_col="PLAYER",
    )
    df_passes["xc"] = ret.xc

    df_tracking_passes = df_tracking[df_tracking["Frame"].isin(df_passes["frame_id"].unique())]
    df_tracking_passes = df_tracking_passes.merge(df_passes[["frame_id", "passer_team"]], right_on="frame_id", left_on="Frame")
    ret_das = accessible_space.interface.get_individual_dangerous_accessible_space(
        df_tracking_passes, frame_col="Frame", x_col="X", y_col="Y", ball_player_id="Ball",
        return_cropped_result=True, player_col="PLAYER", team_col="TEAM", team_in_possession_col="passer_team",
        period_col="Period",
    )
    df_tracking_passes["AS"] = ret_das.acc_space
    df_tracking_passes["DAS"] = ret_das.das
    df_tracking_passes["frame_index"] = ret_das.frame_index
    df_tracking_passes["AS_player"] = ret_das.player_acc_space
    df_tracking_passes["DAS_player"] = ret_das.player_das

    for team, df_tracking_passes_team in df_tracking_passes.groupby("passer_team"):
        df_tracking_passes_team = df_tracking_passes_team[df_tracking_passes_team["TEAM"] == team]
        assert df_tracking_passes_team.groupby("Frame").agg({"DAS": "nunique"})["DAS"].max() == 1
        assert df_tracking_passes_team.groupby("Frame").agg({"AS": "nunique"})["AS"].max() == 1

        ### TODO fix this test
        # dfg = df_tracking_passes_team.groupby("Frame").agg({"DAS": "first", "AS": "first", "DAS_player": "sum", "AS_player": "sum"})
        # st.write("dfg")
        # st.write(dfg)
        # st.write(dfg[["DAS", "DAS_player"]])
        # i_path = dfg["DAS"] > dfg["DAS_player"]
        # st.write("i_path")
        # st.write(i_path)
        # st.write(dfg[i_path])
        # assert (dfg["DAS"] <= dfg["DAS_player"]).all()  # DAS/AS is non-exclusive within each team
        # assert (dfg["AS"] <= dfg["AS_player"]).all()

    df_passes["DAS_from_das"] = df_passes["frame_id"].map(df_tracking_passes.set_index("frame_id")["DAS"].to_dict())
    df_passes["AS_from_das"] = df_passes["frame_id"].map(df_tracking_passes.set_index("frame_id")["AS"].to_dict())
    df_passes["frame_index"] = df_passes["frame_id"].map(df_tracking_passes.set_index("frame_id")["frame_index"].to_dict())
    # st.write("df_passes")
    # st.write(df_passes)

    # st.write("Plotting xC")
    for pass_nr, (pass_index, p4ss) in enumerate(df_passes.iloc[:30].iterrows()):
        plt.figure()
        plt.arrow(p4ss["coordinates_x"], p4ss["coordinates_y"], p4ss["end_coordinates_x"]-p4ss["coordinates_x"], p4ss["end_coordinates_y"]-p4ss["coordinates_y"], color="black", head_width=2, head_length=3)
        df_tracking_frame = df_tracking[df_tracking["Frame"] == p4ss["frame_id"]]
        for team, df_tracking_frame_team in df_tracking_frame.groupby("TEAM"):
            team2color = {"Home": "red", "Away": "blue"}
            plt.scatter(df_tracking_frame_team["X"], df_tracking_frame_team["Y"], color=team2color[team])

        # df_tracking_frame["team_in_possession"] = p4ss["passer_team"]
        # ret = accessible_space.get_dangerous_accessible_space(
        #     df_tracking_frame, frame_col="Frame", x_col="X", y_col="Y", ball_player_id="Ball",
        #     return_cropped_result=True, player_col="PLAYER", team_col="TEAM",
        #
        #     **parameters,
        # )
        plt.xlim([-52.5, 52.5])
        plt.ylim([-34, 34])

        plt.title(f"Frame: {p4ss['frame_id']}, xC: {p4ss['xc']:1f}, Success: {p4ss['is_successful']}, DAS: {p4ss['DAS_from_das']}, AS: {p4ss['AS_from_das']}")

        accessible_space.plot_expected_completion_surface(ret_das.simulation_result, p4ss["frame_index"])
        # st.write(plt.gcf())
        # df_passes.loc[pass_index, "DAS_from_das"] = ret.das.iloc[0]
        # df_passes.loc[pass_index, "AS_from_das"] = ret.acc_space.iloc[0]

        plt.close()

    # average_xc = df_passes["xc"].mean()
    # average_success = df_passes["success"].mean()
    # st.write("xc", average_xc, "success", average_success)
    # assert average_xc > 0.5
    # assert average_success < 1.0

    def calculate_log_loss(y_true, y_pred):
        epsilon = 1e-15  # Avoid log(0) errors
        y_pred = np.clip(y_pred, epsilon, 1 - epsilon)  # Clip probabilities to avoid extreme values
        log_loss = -np.mean(y_true * np.log(y_pred) + (1 - y_true) * np.log(1 - y_pred))
        return log_loss

    # logloss = sklearn.metrics.log_loss(df_passes["is_successful"], df_passes["xc"], labels=np.array([0, 1]))
    logloss = calculate_log_loss(df_passes["is_successful"], df_passes["xc"])
    # st.write("logloss", logloss)
    # brier = sklearn.metrics.brier_score_loss(df_passes["is_successful"], df_passes["xc"])
    brier = np.mean((df_passes["is_successful"] - df_passes["xc"]) ** 2)
    # st.write("brier", brier)

    assert logloss < 0.6
    assert brier < 0.2

    # @st.cache_resource
    def _get2():
        ret2 = accessible_space.get_das_gained(
            df_passes, df_tracking,
            tracking_frame_col="Frame",
            tracking_x_col="X",
            tracking_y_col="Y",
            tracking_team_col="TEAM",
            tracking_player_col="PLAYER",
            tracking_period_col="Period",
            ball_tracking_player_id="Ball",
            event_frame_col="frame_id",
            event_target_frame_col="end_frame_id",
            event_success_col="is_successful",
            event_team_col="passer_team",
            event_receiver_team_col="receiver_team",
            # event_start_x_col="Start X",
            # event_start_y_col="Start Y",
            # event_target_x_col="End X",
            # event_target_y_col="End Y",
            use_event_coordinates_as_ball_position=False,
            use_event_team_as_team_in_possession=True,
            use_progress_bar=True,
        )
        return ret2

    ret2 = _get2()

    df_passes["DAS_gained"] = ret2.das_gained
    df_passes["AS_gained"] = ret2.as_gained
    df_passes["AS_from_gained"] = ret2.acc_space
    df_passes["AS_reception"] = ret2.acc_space_reception
    df_passes["DAS_from_gained"] = ret2.das
    df_passes["DAS_reception"] = ret2.das_reception
    df_passes["frame_index"] = ret2.frame_index
    # st.write("df_passes")
    # st.write(df_passes)

    # sort by das gained
    df_best_passes = df_passes.sort_values("DAS_gained", ascending=False)
    df_worst_passes = df_passes.sort_values("DAS_gained", ascending=True)

    # plot most dangerous passes
    for (df, string) in [
        (df_best_passes, "best"),
        (df_worst_passes, "worst"),
    ]:
        # st.write(f"### {string} passes")
        for _, p4ss in df.iloc[:30].iterrows():
            plt.figure()
            plt.arrow(p4ss["coordinates_x"], p4ss["coordinates_y"], p4ss["end_coordinates_x"]-p4ss["coordinates_x"], p4ss["end_coordinates_y"]-p4ss["coordinates_y"], color="black", head_width=2, head_length=3)
            df_tracking_frame = df_tracking[df_tracking["Frame"] == p4ss["frame_id"]]
            for team, df_tracking_frame_team in df_tracking_frame.groupby("TEAM"):
                team2color = {"Home": "red", "Away": "blue"}
                plt.scatter(df_tracking_frame_team["X"], df_tracking_frame_team["Y"], color=team2color[team])

            plt.xlim([-52.5, 52.5])
            plt.ylim([-34, 34])
            plt.title(f"Frame: {p4ss['frame_id']}, xC: {p4ss['xc']:1f}, Success: {p4ss['is_successful']}, DAS gained: {p4ss['DAS_gained']}, AS gained: {p4ss['AS_gained']}")

            accessible_space.plot_expected_completion_surface(ret2.simulation_result, p4ss["frame_index"])

            # st.write(plt.gcf())
            plt.close()

    df_passes["diff1"] = df_passes["DAS_from_das"] - df_passes["DAS_from_gained"]
    df_passes["diff2"] = df_passes["AS_from_das"] - df_passes["AS_from_gained"]

    # st.write(df_passes[['DAS_from_das', 'DAS_from_gained', 'AS_from_das', 'AS_from_gained', "diff1", "diff2"]])

    # check all close
    assert np.mean(np.isclose(df_passes["DAS_from_das"], df_passes["DAS_from_gained"])) >= 0.9  # DAS and DAS Gained sometimes don't match when team in possession is not clear for a frame.
    assert np.mean(np.isclose(df_passes["AS_from_das"], df_passes["AS_from_gained"])) >= 0.9
